"""
This module provides functions to work in the special orthogonal group SO(3) and its Lie algebra so(3).

Functions:
    rot_inv(rot): Compute the inverse of a rotation matrix.
    hpr2quat(hpr): Convert Heading, Pitch and Roll (HPR) angles to quaternion using the ZYX Euler angles convention.
    rph2quat(rph): Convert Roll, Pitch, Heading (RPH) angles to quaternion using the ZYX Euler angles convention.
    quat2hpr(quat): Convert quaternion to Heading, Pitch and Roll (HPR) angles using the ZYX Euler angles convention.
    quat2rph(quat): Convert quaternion to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles convention.
    hpr2rot(hpr): Convert Heading, Pitch and Roll (HPR) angles to SO(3) rotation matrix using the ZYX Euler angles
                  convention.
    rph2rot(rph): Convert Roll, Pitch, Heading (RPH) angles to SO(3) rotation matrix using the ZYX Euler angles
                  convention.
    quat2rot(quat): Convert quaternion to a rotation matrix in SO(3) using the ZYX Euler angles convention.
    rot2hpr(rot): Convert SO(3) rotation matrix to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles
                  convention.
    rot2rph(rot): Convert SO(3) rotation matrix to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles
                  convention.
    rot2quat(rot): Convert SO(3) rotation matrix to quaternion using the ZYX Euler angles convention.
    rot_diff(rot_1, rot_2): Compute the relative rotation between two matrices.
    vec_to_so3(vec): Compute the skew-symmetric representation for a vector of R3.
    so3_to_vec(mat): Convert a so3 matrix into a R3 vector.
    axis_ang3(vec): Converts a R3 representing a rotation into a exponential coordinates parametrization.
    matrix_exp3(mat): Compute the matrix exponential of a matrix in so(3) representing the exponential coordinates
                      for rotation.
    matrix_log3(mat): Compute the matrix logarithm of a matrix in SO(3) representing the exponential coordinates
                        for rotation.
    project_to_so3(mat): Project a matrix in R3 to SO(3).
    distance_to_so3(mat_1, mat_2): Compute the distance between two matrices in SO(3).
    left_quat_product_matrix(quat): Compute the left quaternion product matrix for a quaternion.
    right_quat_product_matrix(quat): Compute the right quaternion product matrix for a quaternion.

Authors: Sebastián Rodríguez-Martínez and Giancarlo Troni
Contact: srodriguez@mbari.org
"""

import math
from typing import Any, Callable, Iterable, Tuple, Union

import numpy as np
import pymanopt


def rot_inv(rot: np.ndarray) -> np.ndarray:
    """
    Compute the inverse of a rotation matrix.

    Args:
        rot (np.ndarray): The rotation matrix as a (3,3) numpy array.

    Returns:
        rot_mat (np.ndarray): The inverse of the rotation matrix as a (3,3) numpy array.

    Raises:
        ValueError: If the input rotation matrix is not a (3,3) numpy array.

    Examples:
        >>> rot_mat = np.array([[0.93629336, -0.27509585, 0.21835066],
                                [0.28962948, 0.95642509, -0.03695701],
                                [-0.19866933, 0.0978434, 0.97517033]])
        >>> rot_inv = rot_inv(rot_mat)
        >>> print(rot_inv)
        [[ 0.93629336  0.28962948 -0.19866933]
         [-0.27509585  0.95642509  0.0978434 ]
         [ 0.21835066 -0.03695701  0.97517033]]

    Notes:
        The input rotation matrix must be orthonormal and have a determinant of +1 to represent a valid rotation.
    """
    if not isinstance(rot, np.ndarray) and not isinstance(rot, list):
        raise ValueError("Usage: The rotation matrix must be a numpy array or a list.")

    rot = np.array(rot) if isinstance(rot, list) else rot

    if rot.shape != (3, 3):
        raise ValueError("Usage: The rotation matrix must have shape (3, 3)")

    return np.einsum("ij -> ji", rot)


def hpr2quat(hpr: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert Heading, Pitch and Roll (HPR) angles to quaternion using the ZYX Euler angles convention.

    Args:
        hpr (Union[np.ndarray, Iterable[float]]): hpr angles in radians as a numpy array or as a list.

    Returns:
        quat (np.ndarray): The quaternion using hpr with the convention ZYX in order (x, y, z, w)

    Raises:
        ValueError: If the input HPR angles are not a 3-element numpy array or list.

    Examples:
        >>> hpr = np.array([0.3, 0.2, 0.1])
        >>> quat = hpr2quat(hpr)
        >>> print(quat)
        [0.93629336 0.28962948 0.19866933 0.0978434 ]

    Notes:
        The HPR angles represent rotations around the Z, Y, and X axes, respectively.
        The resulting quaternion represents a rotation of the input HPR angles using the ZYX Euler angles convention.
    """
    hpr = np.array(hpr).flatten() if isinstance(hpr, list) else hpr.flatten()
    if hpr.shape[0] != 3:
        raise ValueError("Usage: quat=hpr2quat([angH angP angR])")

    angH = hpr[0]
    angP = hpr[1]
    angR = hpr[2]

    ch = math.cos(angH / 2)
    sh = math.sin(angH / 2)
    cp = math.cos(angP / 2)
    sp = math.sin(angP / 2)
    cr = math.cos(angR / 2)
    sr = math.sin(angR / 2)

    quat = np.array(
        [
            ch * cp * sr - sh * sp * cr,
            ch * sp * cr + sh * cp * sr,
            sh * cp * cr - ch * sp * sr,
            ch * cp * cr + sh * sp * sr,
        ]
    )

    return quat / np.linalg.norm(quat)


def rph2quat(rph: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert Roll, Pitch, Heading (RPH) angles to quaternion using the ZYX Euler angles convention.

    Args:
        rph (Union[np.ndarray, Iterable[float]]): rph angles in radians as a numpy array or as a list.

    Returns:
        quat (np.ndarray): The quaternion using rph with the convention ZYX in order (x, y, z, w)

    Raises:
        ValueError: If the input RPH angles are not a 3-element numpy array or list.

    Examples:
        >>> rph = np.array([0.1, 0.2, 0.3])
        >>> quat = rph2quat(rph)
        >>> print(quat)
        [0.93629336 0.28962948 0.19866933 0.0978434 ]

    Notes:
        The RPH angles represent rotations around the X, Y, and Z axes, respectively.
        The resulting quaternion represents a rotation of the input RPH angles using the ZYX Euler angles convention.
    """
    rph = np.array(rph).flatten() if isinstance(rph, list) else rph.flatten()
    if rph.shape[0] != 3:
        raise ValueError("Usage: quat=rph2quat([angH angP angR])")

    angR = rph[0]
    angP = rph[1]
    angH = rph[2]

    ch = math.cos(angH / 2)
    sh = math.sin(angH / 2)
    cp = math.cos(angP / 2)
    sp = math.sin(angP / 2)
    cr = math.cos(angR / 2)
    sr = math.sin(angR / 2)

    quat = np.array(
        [
            ch * cp * sr - sh * sp * cr,
            ch * sp * cr + sh * cp * sr,
            sh * cp * cr - ch * sp * sr,
            ch * cp * cr + sh * sp * sr,
        ]
    )

    return quat / np.linalg.norm(quat)


def quat2hpr(quat: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert quaternion to Heading, Pitch and Roll (HPR) angles using the ZYX Euler angles convention.

    Args:
        quat (Union[np.ndarray, Iterable[float]]): The quaternion as a numpy array or as a list in the order
                                                    (x, y, z, w).

    Returns:
        hpr (np.ndarray): The HPR angles in radians as a numpy array with the convention ZYX.

    Raises:
        ValueError: If the input quaternion is not a 4-element numpy array or list.
        ValueError: If the input quaternion is not normalized.

    Examples:
        >>> quat = np.array([0.93629336, 0.28962948, 0.19866933, 0.0978434])
        >>> hpr = quat2hpr(quat)
        >>> print(hpr)
        [0.3 0.2 0.1]

    Notes:
        The resulting HPR angles represent rotations around the Z, Y, and X axes, respectively.
        The input quaternion must be normalized to represent a valid rotation.
    """
    quat = np.array(quat).flatten() if isinstance(quat, list) else quat.flatten()
    if quat.shape[0] != 4:
        raise ValueError("Usage: hpr=quat2hpr([x y z w])")

    if abs(np.linalg.norm(quat) - 1.0) >= 1e-6:
        raise ValueError("Usage: The quaternion must be normalized.")

    q0 = quat[3]
    q1 = quat[0]
    q2 = quat[1]
    q3 = quat[2]

    angH = math.atan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2 * q2 + q3 * q3))
    angP = math.asin(2 * (q0 * q2 - q3 * q1))
    angR = math.atan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1 * q1 + q2 * q2))

    return np.array([angH, angP, angR])


def quat2rph(quat: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert quaternion to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles convention.

    Args:
        quat (Union[np.ndarray, Iterable[float]]): The quaternion as a numpy array or as a list in the order
                                                    (x, y, z, w).

    Returns:
        rph (np.ndarray): The RPH angles in radians as a numpy array with the convention ZYX.

    Raises:
        ValueError: If the input quaternion is not a 4-element numpy array or list.
        ValueError: If the input quaternion is not normalized.

    Examples:
        >>> quat = np.array([0.93629336, 0.28962948, 0.19866933, 0.0978434])
        >>> rph = quat2rph(quat)
        >>> print(rph)
        [0.1 0.2 0.3]

    Notes:
        The resulting RPH angles represent rotations around the X, Y, and Z axes, respectively.
        The input quaternion must be normalized to represent a valid rotation.
    """
    quat = np.array(quat).flatten() if isinstance(quat, list) else quat.flatten()
    if quat.shape[0] != 4:
        raise ValueError("Usage: rph=quat2rph([x y z w])")

    if abs(np.linalg.norm(quat) - 1.0) >= 1e-6:
        raise ValueError("Usage: The quaternion must be normalized.")

    q0 = quat[3]
    q1 = quat[0]
    q2 = quat[1]
    q3 = quat[2]

    angR = math.atan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1 * q1 + q2 * q2))
    angP = math.asin(2 * (q0 * q2 - q3 * q1))
    angH = math.atan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2 * q2 + q3 * q3))

    return np.array([angR, angP, angH])


def hpr2rot(hpr: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert Heading, Pitch and Roll (HPR) angles to SO(3) rotation matrix using the ZYX Euler angles convention.

    Args:
        hpr (Union[np.ndarray, Iterable[float]]): hpr angles in radians as a numpy array or as a list.

    Returns:
        rot (np.ndarray): The rotation matrix using hpr with the convention ZYX

    Raises:
        ValueError: If the input HPR angles are not a 3-element numpy array or list.

    Examples:
        >>> hpr = np.array([0.3, 0.2, 0.1])
        >>> rot_mat = hpr2rot(hpr)
        >>> print(rot_mat)
        [[ 0.93629336 -0.27509585  0.21835066]
         [ 0.28962948  0.95642509 -0.03695701]
         [-0.19866933  0.0978434   0.97517033]]

    Notes:
        The HPR angles represent rotations around the Z, Y, and X axes, respectively.
        The resulting rotation matrix represents a rotation of the input HPR angles using the ZYX Euler angles
        convention.
    """
    hpr = np.array(hpr).flatten() if isinstance(hpr, list) else hpr.flatten()
    if hpr.shape[0] != 3:
        raise ValueError("Usage: RotMat=hpr2rot([angH angP angR])")

    angH = hpr[0]
    angP = hpr[1]
    angR = hpr[2]

    c1 = math.cos(angH)
    s1 = math.sin(angH)
    c2 = math.cos(angP)
    s2 = math.sin(angP)
    c3 = math.cos(angR)
    s3 = math.sin(angR)

    RotMat = np.array(
        [
            [c1 * c2, c1 * s2 * s3 - c3 * s1, s1 * s3 + c1 * c3 * s2],
            [c2 * s1, c1 * c3 + s1 * s2 * s3, c3 * s1 * s2 - c1 * s3],
            [-s2, c2 * s3, c2 * c3],
        ]
    )

    return RotMat


def rph2rot(rph: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert Roll, Pitch, Heading (RPH) angles to SO(3) rotation matrix using the ZYX Euler angles convention.

    Args:
        rph (Union[np.ndarray, Iterable[float]]): rph angles in radians as a numpy array or as a list.

    Returns:
        rot (np.ndarray): The rotation matrix using rph with the convention ZYX

    Raises:
        ValueError: If the input RPH angles are not a 3-element numpy array or list.

    Examples:
        >>> rph = np.array([0.1, 0.2, 0.3])
        >>> rot_mat = rph2rot(rph)
        >>> print(rot_mat)
        [[ 0.93629336 -0.27509585  0.21835066]
         [ 0.28962948  0.95642509 -0.03695701]
         [-0.19866933  0.0978434   0.97517033]]

    Notes:
        The RPH angles represent rotations around the X, Y, and Z axes, respectively.
        The resulting rotation matrix represents a rotation of the input RPH angles using the ZYX Euler angles
        convention.
    """
    rph = np.array(rph).flatten() if isinstance(rph, list) else rph.flatten()
    if rph.shape[0] != 3:
        raise ValueError("Usage: RotMat=rph2rot([angH angP angR])")

    angR = rph[0]
    angP = rph[1]
    angH = rph[2]

    c1 = math.cos(angH)
    s1 = math.sin(angH)
    c2 = math.cos(angP)
    s2 = math.sin(angP)
    c3 = math.cos(angR)
    s3 = math.sin(angR)

    RotMat = np.array(
        [
            [c1 * c2, c1 * s2 * s3 - c3 * s1, s1 * s3 + c1 * c3 * s2],
            [c2 * s1, c1 * c3 + s1 * s2 * s3, c3 * s1 * s2 - c1 * s3],
            [-s2, c2 * s3, c2 * c3],
        ]
    )

    return RotMat


def quat2rot(quat: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Convert quaternion to a rotation matrix in SO(3) using the ZYX Euler angles convention.

    Args:
        quat (Union[np.ndarray, Iterable[float]]): The quaternion as a numpy array or as a list in the order
                                                    (x, y, z, w).

    Returns:
        rot (np.ndarray): The RPH angles in radians as a numpy array with the convention ZYX.

    Raises:
        ValueError: If the input quaternion is not a 4-element numpy array or list.
        ValueError: If the input quaternion is not normalized.

    Examples:
        >>> quat = np.array([0.93629336, 0.28962948, 0.19866933, 0.0978434])
        >>> rotmat = quat2rot(quat)
        >>> print(rotmat)
        [[ 0.93629336 -0.27509585  0.21835066]
         [ 0.28962948  0.95642509 -0.03695701]
         [-0.19866933  0.0978434   0.97517033]]

    Notes:
        The input quaternion must be normalized to represent a valid rotation.
    """
    quat = np.array(quat).flatten() if isinstance(quat, list) else quat.flatten()
    if quat.shape[0] != 4:
        raise ValueError("Usage: rph=quat2rph([x y z w])")

    if abs(np.linalg.norm(quat) - 1.0) >= 1e-6:
        raise ValueError("Usage: The quaternion must be normalized.")

    rph = quat2rph(quat)
    rot = rph2rot(rph)

    return rot


def rot2hpr(rot: np.ndarray) -> np.ndarray:
    """
    Convert SO(3) rotation matrix to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles convention.

    Args:
        rot (np.ndarray): The rotation matrix as a (3,3) numpy array.

    Returns:
        hpr (np.ndarray): The RPH angles in radians as a numpy array with the convention ZYX.

    Raises:
        ValueError: If the input rotation matrix is not a (3,3) numpy array or list.

    Examples:
        >>> rot_mat = np.array([[0.93629336, -0.27509585, 0.21835066],
                                [0.28962948, 0.95642509, -0.03695701],
                                [-0.19866933, 0.0978434, 0.97517033]])
        >>> hpr = rot2hpr(rot_mat)
        >>> print(hpr)
        [0.3 0.2 0.1]

    Notes:
        The resulting RPH angles represent rotations around the X, Y, and Z axes, respectively.
        The input rotation matrix must be orthonormal and have a determinant of +1 to represent a valid rotation.
    """
    if (not isinstance(rot, np.ndarray)) and (not isinstance(rot, list)):
        raise ValueError("Usage: The rotation matrix must be a numpy array or a list.")

    rot = np.array(rot) if isinstance(rot, list) else rot

    if rot.shape != (3, 3):
        raise ValueError("Usage: The rotation matrix must have shape (3, 3)")

    h = math.atan2(rot[1, 0], rot[0, 0])
    p = math.atan2(
        -rot[2, 0], math.sqrt(math.pow(rot[2, 1], 2) + math.pow(rot[2, 2], 2))
    )
    r = math.atan2(rot[2, 1], rot[2, 2])

    return np.array([h, p, r])


def rot2rph(rot: np.ndarray) -> np.ndarray:
    """
    Convert SO(3) rotation matrix to Roll, Pitch, Heading (RPH) angles using the ZYX Euler angles convention.

    Args:
        rot (np.ndarray): The rotation matrix as a (3,3) numpy array.

    Returns:
        rph (np.ndarray): The RPH angles in radians as a numpy array with the convention ZYX.

    Raises:
        ValueError: If the input rotation matrix is not a (3,3) numpy array.

    Examples:
        >>> rot_mat = np.array([[0.93629336, -0.27509585, 0.21835066],
                                [0.28962948, 0.95642509, -0.03695701],
                                [-0.19866933, 0.0978434, 0.97517033]])
        >>> rph = rot2rph(rot_mat)
        >>> print(rph)
        [0.1 0.2 0.3]

    Notes:
        The resulting RPH angles represent rotations around the X, Y, and Z axes, respectively.
        The input rotation matrix must be orthonormal and have a determinant of +1 to represent a valid rotation.
    """
    if (not isinstance(rot, np.ndarray)) and (not isinstance(rot, list)):
        raise ValueError("Usage: The rotation matrix must be a numpy array or a list.")

    rot = np.array(rot) if isinstance(rot, list) else rot

    if rot.shape != (3, 3):
        raise ValueError("Usage: The rotation matrix must have shape (3, 3)")

    h = math.atan2(rot[1, 0], rot[0, 0])
    p = math.atan2(
        -rot[2, 0], math.sqrt(math.pow(rot[2, 1], 2) + math.pow(rot[2, 2], 2))
    )
    r = math.atan2(rot[2, 1], rot[2, 2])

    return np.array([r, p, h])


def rot2quat(rot: np.ndarray) -> np.ndarray:
    """
    Convert SO(3) rotation matrix to quaternion using the ZYX Euler angles convention.

    Args:
        rot (np.ndarray): The rotation matrix as a (3,3) numpy array.

    Returns:
        quat (np.ndarray): The quaternion as a numpy array with the convention ZYX in order (x, y, z, w)

    Raises:
        ValueError: If the input rotation matrix is not a (3,3) numpy array.

    Examples:
        >>> rot_mat = np.array([[0.50000000, -0.1464466,  0.8535534],
                                [0.50000000,  0.8535534, -0.1464466],
                                [-0.7071068,  0.5000000,  0.5000000]])
        >>> quat = rot2quat(rot_mat)
        >>> print(quat)
        [0.1913417, 0.4619398, 0.1913417, 0.8446232]

    Notes:
        The input rotation matrix must be orthonormal and have a determinant of +1 to represent a valid rotation.
    """
    if (not isinstance(rot, np.ndarray)) and (not isinstance(rot, list)):
        raise ValueError("Usage: The rotation matrix must be a numpy array or a list.")

    rot = np.array(rot) if isinstance(rot, list) else rot

    if rot.shape != (3, 3):
        raise ValueError("Usage: The rotation matrix must have shape (3, 3)")

    rph = rot2rph(rot)
    quat = rph2quat(rph)

    return quat / np.linalg.norm(quat)


def rot_diff(rot_1: np.ndarray, rot_2: np.ndarray) -> np.ndarray:
    """
    Compute the relative rotation between two matrices.

    The returned matrix is the attitude of 2 with respect to 1:

    $$ R_{12} = R_{1w} \\cdot R_{w2} = R_{w1}^{-1} \\cdot R_{w2} $$

    Args:
        rot_1 (np.ndarray): Rotation matrix for attitude 1 as a numpy array.
        rot_2 (np.ndarray): Rotation matrix for attitude 2 as a numpy array.

    Returns:
        rot_1_2 (np.ndarray): The attitude of 2 with respect to 1 as a numpy array.

    Raises:
        ValueError: If the input rotation matrices are not (3,3) numpy arrays.

    Examples:
        >>> rot_mat_1 = np.array([[0.93629336, -0.27509585, 0.21835066],
                                [0.28962948, 0.95642509, -0.03695701],
                                [-0.19866933, 0.0978434, 0.97517033]])
        >>> rot_mat_2 = np.array([[0.82298387, -0.41998535, 0.38175244],
                                [0.56975033, 0.81758362, -0.08209422],
                                [-0.0, 0.3939193, 0.91914503]])
        >>> rot_diff = rotDiff(rot_mat_1, rot_mat_2)
        >>> print(rot_diff)
        [[ 0.93557083, -0.23469286,  0.15104944],
         [ 0.31852406,  0.93603611, -0.09360322],
         [ 0.1586428 ,  0.26221888,  0.98271281]]

    Notes:
        The input rotation matrices must be orthonormal and have a determinant of +1 to represent valid rotations.
    """
    if not isinstance(rot_1, np.ndarray) or not isinstance(rot_2, np.ndarray):
        raise ValueError("Usage: The rotation matrices must be numpy arrays.")
    if rot_1.shape != (3, 3) or rot_2.shape != (3, 3):
        raise ValueError("Usage: The rotation matrix must have shape (3, 3)")
    return np.einsum("ij -> ji", rot_1) @ rot_2


def vec_to_so3(vec: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Compute the skew-symmetric representation for a vector of $\\mathbf{R}^3$.

    $$ \\lfloor x \\rfloor =
    \\begin{bmatrix}
    0 & -z & y \\\\
    z & 0 & -x \\\\
    -y & x & 0
    \\end{bmatrix}
    $$

    Args:
        vec (Union[np.ndarray, Iterable[float]]): The vector as a numpy array or a list.

    Returns:
        skew (np.ndarray): The skew-symmetric matrix as a numpy array.

    Raises:
        ValueError: If the input vector is not a numpy array or a list.
        ValueError: If the input vector does not have 3 components.

    Examples:
        >>> vec = np.array([1, 2, 3])
        >>> skew_sym = vec_to_so3(vec)
        >>> print(skew_sym)
        [[ 0. -3.  2.]
         [ 3.  0. -1.]
         [-2.  1.  0.]]
    """
    if (not isinstance(vec, np.ndarray)) and (not isinstance(vec, list)):
        raise ValueError("Type Error: The vector must be a numpy array or a list.")
    if isinstance(vec, np.ndarray):
        vec = vec.flatten()
    if len(vec) != 3:
        raise ValueError("Shape Error: The vector must have only 3 components.")
    matrix = np.zeros((3, 3))
    matrix[0, 1], matrix[0, 2], matrix[1, 2] = -vec[2], vec[1], -vec[0]
    return matrix - matrix.T


def so3_to_vec(mat: np.ndarray) -> np.ndarray:
    """
    Convert a so3 matrix into a $\\mathbf{R}^3$ vector.

    Args:
        mat (np.ndarray): The skew-symmetric matrix as a (3,3) numpy array.

    Returns:
        vec (np.ndarray): The vector as a (3, ) numpy array.

    Raises:
        ValueError: If the input matrix is not a (3,3) numpy array.

    Examples:
        >>> skew_sym = np.array([[0, -3, 2], [3, 0, -1], [-2, 1, 0]])
        >>> vec = so3_to_vec(skew_sym)
        >>> print(vec)
        [1. 2. 3.]
    """
    if not isinstance(mat, np.ndarray) and not isinstance(mat, list):
        raise ValueError("Type Error: The matrix must be a numpy array or a list.")

    mat = np.array(mat) if isinstance(mat, list) else mat

    if mat.shape != (3, 3):
        raise ValueError("Shape Error: The matrix must have shape (3, 3).")

    return np.array([mat[2][1], mat[0][2], mat[1][0]])


def axis_ang3(vec: Union[Iterable[float], np.ndarray]) -> Tuple[np.ndarray, float]:
    """
    Converts a $\\mathbf{R}^3$ representing a rotation into a exponential coordinates
    parametrization, i.e., a unit vector representing the axis (w) and an
    angle theta.

    Args:
        vec (Union[Iterable[float], np.ndarray]): R3 vector representing a rotation in so3.

    Returns:
        axis (np.ndarray): Unit vector representing the axis of rotation as a numpy array.
        angle (float): Angle of rotation in radians.

    Raises:
        ValueError: If the input vector is not a numpy array or a list.
        ValueError: If the input vector does not have 3 components.

    Examples:
        >>> vec = np.array([1, 2, 3])
        >>> axis, angle = axis_ang_3(vec)
        >>> print(axis)
        [0.26726124 0.53452248 0.80178373]
        >>> print(angle)
        3.7416573867739413
    """
    if (not isinstance(vec, np.ndarray)) and (not isinstance(vec, list)):
        raise ValueError("Type Error: The vector must be a numpy array or a list.")

    vec = np.array(vec).flatten() if isinstance(vec, list) else vec.flatten()

    if len(vec) != 3:
        raise ValueError("Shape Error: The vector must have only 3 components.")

    if np.allclose(vec, np.zeros(3)):
        return np.zeros(3), 0

    vec_norm = np.linalg.norm(vec)
    return vec / vec_norm, vec_norm


def matrix_exp3(mat: np.ndarray) -> np.ndarray:
    """
    Compute the matrix exponential of a matrix in so(3) representing the
    exponential coordinates for rotation, i.e., [w]*theta, where [w] is the
    skew-symmetric matrix of the unitary rotation axis. To compute the exponential
    matrix, the Rodrigues' formula is used.

    Args:
        mat (np.ndarray): Skew-symmetric matrix representing the exponential coordinates of rotation as a (3, 3)
            numpy array.

    Returns:
        mat_exp3 (np.ndarray): Matrix exponential of the exponential coordinates as a numpy array.

    Raises:
        ValueError: If the input matrix is not a (3,3) numpy array.
        ValueError: If the input matrix is not a skew-symmetric matrix.

    Examples:
        >>> skew_sym = np.array([[0, -3, 2], [3, 0, -1], [-2, 1, 0]])
        >>> exp_mat = matrix_exp_3(skew_sym)
        >>> print(exp_mat)
        [[-0.69492056  0.71352099  0.08929286]
         [-0.19200697 -0.30378504  0.93319235]
         [ 0.69297817  0.63134970   0.34810748]]
    """
    if not isinstance(mat, np.ndarray) and not isinstance(mat, list):
        raise ValueError("Type Error: The matrix must be a numpy array or a list.")

    mat = np.array(mat) if isinstance(mat, list) else mat

    if mat.ndim != 2:
        raise ValueError("Shape Error: The matrix must be a 2-dimensional numpy array.")
    if mat.shape != (3, 3):
        raise ValueError("Shape Error: The matrix must be a (3, 3) numpy array.")

    rotation = so3_to_vec(mat)

    # Check if the rotation is zero to avoid singularity
    if np.abs(np.linalg.norm(rotation)) < 1.00e-06:
        return np.eye(3, dtype=np.float64)

    else:
        theta = axis_ang3(rotation)[1]
        unit_mat = mat / theta
        return (
            np.eye(3, dtype=np.float64)
            + np.sin(theta) * unit_mat
            + (1 - np.cos(theta)) * (unit_mat @ unit_mat)
        )


def matrix_log3(R: np.ndarray) -> np.ndarray:
    """
    Computes the matrix logarithm of a rotation matrix.

    Args:
        R (np.ndarray): Rotation matrix as a (3, 3) numpy array.

    Returns:
        mat_log3 (np.ndarray): Matrix logarithm of the rotation matrix as a numpy array.

    Raises:
        ValueError: If the input matrix is not a (3,3) numpy array.
        ValueError: If the input matrix is not a rotation matrix.

    Examples:
        >>> rot_mat = np.array([[0.93629336, -0.27509585, 0.21835066],
                                [0.28962948, 0.95642509, -0.03695701],
                                [-0.19866933, 0.0978434, 0.97517033]])
        >>> log_mat = matrix_log_3(rot_mat)
        >>> print(log_mat)
        [[ 0.        , -0.28874894,  0.21322592],
         [ 0.28874894,  0.        , -0.06892461],
         [-0.21322592,  0.06892461,  0.        ]]
    """
    if not isinstance(R, np.ndarray) and not isinstance(R, list):
        raise ValueError("Type Error: The matrix must be a numpy array or a list.")

    R = np.array(R) if isinstance(R, list) else R

    if R.ndim != 2:
        raise ValueError("Shape Error: The matrix must be a 2-dimensional numpy array.")
    if R.shape != (3, 3):
        raise ValueError("Shape Error: The matrix must be a (3, 3) numpy array.")

    acos_arg = (np.trace(R) - 1) * 0.5

    # acos is undefined
    if acos_arg >= 1:
        return np.zeros((3, 3))

    # if theta = pi, three possible solutions.
    elif acos_arg <= -1:
        if not (np.abs(1 + R[2, 2]) < 1.00e-06):
            omg = (1.0 / np.sqrt(2 * (1 + R[2, 2]))) * (
                R[:, 2] + np.array([0.0, 0.0, 1.0])
            )
        elif not (np.abs(1 + R[1, 1]) < 1.00e-06):
            omg = (1.0 / np.sqrt(2 * (1 + R[1, 1]))) * (
                R[:, 1] + np.array([0.0, 1.0, 0.0])
            )
        else:
            omg = (1.0 / np.sqrt(2 * (1 + R[0, 0]))) * (
                R[:, 0] + np.array([1.0, 0.0, 0.0])
            )
        return vec_to_so3(np.pi * omg)

    else:
        theta = np.arccos(acos_arg)
        return theta * (1 / (2.0 * np.sin(theta))) * (R - R.T)


def project_to_so3(mat: np.ndarray) -> np.ndarray:
    """
    Projects a matrix to the closest matrix in SO(3) using singular value
    decomposition.
    Source: http://hades.mech.northwestern.edu/index.php/Modern_Robotics_Linear_Algebra_Review

    Args:
        mat (np.ndarray): The matrix as a (3,3) numpy array.

    Returns:
        mat_so3 (np.ndarray): The closest matrix in SO(3) as a numpy array.

    Raises:
        ValueError: If the input matrix is not a (3,3) numpy array.

    Examples:
        >>> mat = np.array([[ 0.675,  0.150,  0.720],
                            [ 0.370,  0.771, -0.511],
                            [-0.630,  0.619,  0.472]])
        >>> proj_mat = projectToSO3(mat)
        >>> print(proj_mat)
        [[ 0.67901136,  0.14894516,  0.71885945],
         [ 0.37320708,  0.77319584, -0.51272279],
         [-0.63218672,  0.61642804,  0.46942137]]
    """
    if not isinstance(mat, np.ndarray) and not isinstance(mat, list):
        raise ValueError("Type Error: The matrix must be a numpy array.")

    mat = np.array(mat) if isinstance(mat, list) else mat

    if mat.ndim != 2:
        raise ValueError("Shape Error: The matrix must be a 2-dimensional numpy array.")
    if mat.shape != (3, 3):
        raise ValueError("Shape Error: The matrix must be a (3, 3) numpy array.")

    U, s, V = np.linalg.svd(mat)
    R = U @ V
    if np.linalg.det(R) < 0:
        # Results may be far from mat.
        R[:, s[2, 2]] = -R[:, s[2, 2]]
    return R


def distance_to_so3(mat: np.ndarray) -> float:
    """
    Compute the frobenius norm to describe the distance of a matrix from
    the SO(3) manifold.

    Args:
        mat (np.ndarray): The matrix as a (3,3) numpy array.

    Returns:
        dist_so3 (float): The frobenius norm as a float.

    Raises:
        ValueError: If the input matrix is not a (3,3) numpy array.

    Examples:
        >>> mat = np.array([[ 0.675,  0.150,  0.720],
                            [ 0.370,  0.771, -0.511],
                            [-0.630,  0.619,  0.472]])
        >>> dist = distanceToSO3(mat)
        >>> print(dist)
        0.0663128178912833
    """
    if not isinstance(mat, np.ndarray) and not isinstance(mat, list):
        raise ValueError("Type Error: The matrix must be a numpy array.")

    mat = np.array(mat) if isinstance(mat, list) else mat

    if mat.ndim != 2:
        raise ValueError("Shape Error: The matrix must be a 2-dimensional numpy array.")
    if mat.shape != (3, 3):
        raise ValueError("Shape Error: The matrix must be a (3, 3) numpy array.")

    if np.linalg.det(mat) > 0:
        return np.linalg.norm(mat.T @ mat - np.eye(3, dtype=np.float64))
    else:
        return 1.00e10


def so3_optimization(
    objective_function: Callable, *args: Any, **kwargs: Any
) -> np.ndarray:
    """
    Perform optimization on the Special Orthogonal Group (SO(3)).
    This function uses the pymanopt library to optimize a given objective
    function on the SO(3) manifold.

    Args:
        objective_function (Callable): The objective function to be optimized.
            If the cost function uses numpy for computations, it MUST use
            the numpy module from autograd, i.e., autograd.numpy. Also, the
            function must return a scalar value.
        *args (Any): Positional arguments to be passed to the objective function.
        **kwargs (Any): Keyword arguments to be passed to the objective function.

    Returns:
        mat_so3_opt (np.ndarray): The optimized SO(3) matrix as a numpy array.

    Raises:
        TypeError: If the objective_function is not callable.
        ValueError: If the optimization does not return a valid SO(3) matrix.
        ValueError: If the result is not a (3,3) numpy array.

    Notes:
        For the cost function you can use `import autograd.numpy as np` to
        use numpy functions.
    """
    # Type check: Ensure objective_function is callable.
    if not callable(objective_function):
        raise TypeError("The objective_function must be callable.")

    manifold = pymanopt.manifolds.SpecialOrthogonalGroup(3)
    cost = pymanopt.function.autograd(manifold)(
        lambda x: objective_function(x, *args, **kwargs)
    )
    problem = pymanopt.Problem(manifold=manifold, cost=cost)
    optimizer = pymanopt.optimizers.TrustRegions(verbosity=0)
    result = optimizer.run(problem).point

    # Check that the result is a (3,3) numpy array.
    if not isinstance(result, np.ndarray) or result.shape != (3, 3):
        raise ValueError(
            "Optimization did not return a valid SO(3) matrix of shape (3, 3)."
        )

    return result


def left_quat_product_matrix(quat: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Compute the left quaternion product matrix for a given quaternion, such that
    q * p = L(q) * p. For a quaternion q = (x, y, z, w), the left quaternion product
    matrix is defined as:

    $$
    L(q) =
    \\begin{bmatrix}
        w & -z & y & x \\\\
        z & w & -x & y \\\\
        -y & x & w & z \\\\
        -x & -y & -z & w
    \\end{bmatrix}
    $$

    Args:
        quat (Union[np.ndarray, Iterable[float]]): The quaternion as a numpy array or as a list in the order
            (x, y, z, w).

    Returns:
        left_quat_product_matrix (np.ndarray): The left quaternion product matrix as a (4, 4) numpy array.

    Raises:
        ValueError: If the input quaternion is not a 4-element numpy array or list.
        ValueError: If the input quaternion is not normalized.
        TypeError: If the input quaternion is not a numpy array or list.
    """
    # Convert to numpy array if input is a list
    if isinstance(quat, list):
        quat = np.array(quat)

    # Check if the input is a numpy array
    if not isinstance(quat, np.ndarray):
        raise TypeError("The quaternion must be a numpy array or a list.")

    # Check if the input quaternion has 4 elements
    quat = quat.squeeze()
    if quat.shape[0] != 4:
        raise ValueError("The quaternion must have 4 elements.")
    if quat.ndim != 1:
        raise ValueError("The quaternion must be a 1-dimensional numpy array or list.")

    # Check if the quaternion is normalized
    if abs(np.linalg.norm(quat) - 1.0) >= 1e-6:
        raise ValueError("The quaternion must be normalized.")

    # Extract quaternion components
    x, y, z, w = quat[0], quat[1], quat[2], quat[3]

    # Construct the left quaternion product matrix
    left_quat_product_matrix = np.array(
        [
            [w, -z, y, x],
            [z, w, -x, y],
            [-y, x, w, z],
            [-x, -y, -z, w],
        ]
    )
    return left_quat_product_matrix


def right_quat_product_matrix(quat: Union[np.ndarray, Iterable[float]]) -> np.ndarray:
    """
    Compute the right quaternion product matrix for a given quaternion, such that
    q * p = R(p) * q For a quaternion q = (x, y, z, w), the right quaternion product
    matrix is defined as:

    $$
    R(q) =
    \\begin{bmatrix}
        w & z & -y & x \\\\
        -z & w & x & y \\\\
        y & -x & w & z \\\\
        -x & -y & -z & w
    \\end{bmatrix}
    $$

    Args:
        quat (Union[np.ndarray, Iterable[float]]): The quaternion as a numpy array or as a list in the order
            (x, y, z, w).

    Returns:
        right_quat_product_matrix (np.ndarray): The right quaternion product matrix as a (4, 4) numpy array.

    Raises:
        ValueError: If the input quaternion is not a 4-element numpy array or list.
        ValueError: If the input quaternion is not normalized.
        TypeError: If the input quaternion is not a numpy array or list.
    """
    # Convert to numpy array if input is a list
    if isinstance(quat, list):
        quat = np.array(quat)

    # Check if the input is a numpy array
    if not isinstance(quat, np.ndarray):
        raise TypeError("The quaternion must be a numpy array or a list.")

    # Check if the input quaternion has 4 elements
    quat = quat.squeeze()
    if quat.shape[0] != 4:
        raise ValueError("The quaternion must have 4 elements.")
    if quat.ndim != 1:
        raise ValueError("The quaternion must be a 1-dimensional numpy array or list.")

    # Check if the quaternion is normalized
    if abs(np.linalg.norm(quat) - 1.0) >= 1e-6:
        raise ValueError("The quaternion must be normalized.")

    # Extract quaternion components
    x, y, z, w = quat[0], quat[1], quat[2], quat[3]

    # Construct the right quaternion product matrix
    right_quat_product_matrix = np.array(
        [
            [w, z, -y, x],
            [-z, w, x, y],
            [y, -x, w, z],
            [-x, -y, -z, w],
        ]
    )
    return right_quat_product_matrix



"""
Test module for the SE3 module.

Authors: Sebastián Rodríguez-Martínez
Contact: srodriguez@mbari.org
Last Update: 2024-03-26
"""

import numpy as np
import pytest

from navlib import (
    adjoint,
    axis_ang6,
    distance_to_se3,
    matrix2xyzrph,
    matrix_exp6,
    matrix_inverse,
    matrix_log6,
    pose_diff,
    project_to_se3,
    se3_to_vec,
    vec_to_se3,
    vec_to_so3,
    xyzrph2matrix,
)
from navlib.math import adjoint as adjoint_relative
from navlib.math import axis_ang6 as axis_ang6_relative
from navlib.math import distance_to_se3 as distance_to_se3_relative
from navlib.math import matrix2xyzrph as matrix2xyzrph_relative
from navlib.math import matrix_exp6 as matrix_exp6_relative
from navlib.math import matrix_inverse as matrix_inverse_relative
from navlib.math import matrix_log6 as matrix_log6_relative
from navlib.math import pose_diff as pose_diff_relative
from navlib.math import project_to_se3 as project_to_se3_relative
from navlib.math import se3_to_vec as se3_to_vec_relative
from navlib.math import vec_to_se3 as vec_to_se3_relative
from navlib.math import vec_to_so3 as vec_to_so3_relative
from navlib.math import xyzrph2matrix as xyzrph2matrix_relative


def test_imports():
    # Checks that the absolute and relative imports work correctly
    assert adjoint_relative.__code__.co_code == adjoint.__code__.co_code
    assert axis_ang6_relative.__code__.co_code == axis_ang6.__code__.co_code
    assert distance_to_se3_relative.__code__.co_code == distance_to_se3.__code__.co_code
    assert matrix2xyzrph_relative.__code__.co_code == matrix2xyzrph.__code__.co_code
    assert matrix_exp6_relative.__code__.co_code == matrix_exp6.__code__.co_code
    assert matrix_inverse_relative.__code__.co_code == matrix_inverse.__code__.co_code
    assert matrix_log6_relative.__code__.co_code == matrix_log6.__code__.co_code
    assert pose_diff_relative.__code__.co_code == pose_diff.__code__.co_code
    assert project_to_se3_relative.__code__.co_code == project_to_se3.__code__.co_code
    assert se3_to_vec_relative.__code__.co_code == se3_to_vec.__code__.co_code
    assert vec_to_se3_relative.__code__.co_code == vec_to_se3.__code__.co_code
    assert vec_to_so3_relative.__code__.co_code == vec_to_so3.__code__.co_code
    assert xyzrph2matrix_relative.__code__.co_code == xyzrph2matrix.__code__.co_code


def test_xyzrph2matrix():
    """
    Test the rph2rot function.
    """
    # Test case 1: Zero rotation
    rph = np.array([0, 0, 0])
    xyz = np.array([1.0, 2.0, 3.0])
    xyzrph = np.concatenate((xyz, rph))
    expected_t_mat = np.eye(4)
    expected_t_mat[:3, 3] = xyz
    assert np.allclose(xyzrph2matrix(xyzrph), expected_t_mat)
    assert np.allclose(xyzrph2matrix(xyzrph.tolist()), expected_t_mat)

    # Test case 2: Rotation around X-axis
    rph = np.array([np.pi / 2, 0, 0])
    xyz = np.array([1.0, 2.0, 3.0])
    xyzrph = np.concatenate((xyz, rph))
    expected_t_mat = np.eye(4)
    expected_t_mat[:3, :3] = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
    expected_t_mat[:3, 3] = xyz
    assert np.allclose(xyzrph2matrix(xyzrph), expected_t_mat)
    assert np.allclose(xyzrph2matrix(xyzrph.tolist()), expected_t_mat)

    # Test case 3: Rotation around Y-axis
    rph = np.array([0, np.pi / 2, 0])
    xyz = np.array([1.0, 2.0, 3.0])
    xyzrph = np.concatenate((xyz, rph))
    expected_t_mat = np.eye(4)
    expected_t_mat[:3, :3] = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    expected_t_mat[:3, 3] = xyz
    assert np.allclose(xyzrph2matrix(xyzrph), expected_t_mat)
    assert np.allclose(xyzrph2matrix(xyzrph.tolist()), expected_t_mat)

    # Test case 4: Rotation around Z-axis
    rph = np.array([0, 0, np.pi / 2])
    xyz = np.array([1.0, 2.0, 3.0])
    xyzrph = np.concatenate((xyz, rph))
    expected_t_mat = np.eye(4)
    expected_t_mat[:3, :3] = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    expected_t_mat[:3, 3] = xyz
    assert np.allclose(xyzrph2matrix(xyzrph), expected_t_mat)
    assert np.allclose(xyzrph2matrix(xyzrph.tolist()), expected_t_mat)

    # Test case 5: Rotation around X, Y, and Z axes
    rph = np.array([np.pi / 2, np.pi / 2, np.pi / 2])
    xyz = np.array([1.0, 2.0, 3.0])
    xyzrph = np.concatenate((xyz, rph))
    expected_t_mat = np.eye(4)
    expected_t_mat[:3, :3] = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    expected_t_mat[:3, 3] = xyz
    assert np.allclose(xyzrph2matrix(xyzrph), expected_t_mat)
    assert np.allclose(xyzrph2matrix(xyzrph.tolist()), expected_t_mat)

    # Test case 6: Wrong input shape
    with pytest.raises(ValueError):
        xyzrph2matrix(np.array([1, 2, 3, 0, 0]))


def test_matrix2xyzrph():
    """
    Test the matrix2xyzrph function.
    """
    # Test case 1: Zero rotation
    xyz = np.array([1.0, 2.0, 3.0])
    t_mat = np.eye(4)
    t_mat[:3, 3] = xyz
    expected_xyzrph = np.concatenate((xyz, np.array([0, 0, 0])))
    assert np.allclose(matrix2xyzrph(t_mat), expected_xyzrph)

    # Test case 2: Rotation around X-axis
    xyz = np.array([1.0, 2.0, 3.0])
    t_mat = np.eye(4)
    t_mat[:3, :3] = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
    t_mat[:3, 3] = xyz
    expected_xyzrph = np.concatenate((xyz, np.array([np.pi / 2, 0, 0])))
    assert np.allclose(matrix2xyzrph(t_mat), expected_xyzrph)

    # Test case 3: Rotation around Y-axis
    xyz = np.array([1.0, 2.0, 3.0])
    t_mat = np.eye(4)
    t_mat[:3, :3] = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    t_mat[:3, 3] = xyz
    expected_xyzrph = np.concatenate((xyz, np.array([0, np.pi / 2, 0])))
    assert np.allclose(matrix2xyzrph(t_mat), expected_xyzrph)

    # Test case 4: Rotation around Z-axis
    xyz = np.array([1.0, 2.0, 3.0])
    t_mat = np.eye(4)
    t_mat[:3, :3] = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    t_mat[:3, 3] = xyz
    expected_xyzrph = np.concatenate((xyz, np.array([0, 0, np.pi / 2])))
    assert np.allclose(matrix2xyzrph(t_mat), expected_xyzrph)

    # Test case 5: Rotation around X, Y, and Z axes
    xyz = np.array([1.0, 2.0, 3.0])
    t_mat = np.eye(4)
    t_mat[:3, :3] = np.array([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
    t_mat[:3, 3] = xyz
    expected_xyzrph = np.concatenate((xyz, np.array([0, np.pi / 2, 0])))
    print(matrix2xyzrph(t_mat))
    print(expected_xyzrph)
    assert np.allclose(matrix2xyzrph(t_mat), expected_xyzrph)

    # Test case 6: Wrong input shape
    with pytest.raises(ValueError):
        matrix2xyzrph(np.eye(3))

    # Test case 7: Wrong input instance
    with pytest.raises(ValueError):
        matrix2xyzrph(np.eye(4).tolist())


def test_invert_matrix():
    """
    Test the matrix_inverse function.
    """
    # Test case 1: Identity matrix
    T = np.eye(4)
    expected_result = np.eye(4)
    np.testing.assert_array_equal(matrix_inverse(T), expected_result)

    # Test case 2: Translation matrix
    T = np.eye(4)
    T[:3, 3] = [1, 2, 3]
    expected_result = np.eye(4)
    expected_result[:3, 3] = [-1, -2, -3]
    np.testing.assert_array_equal(matrix_inverse(T), expected_result)

    # Test case 3: Rotation matrix
    T = np.eye(4)
    T[:3, :3] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]  # 90 degree rotation around Z-axis
    expected_result = np.eye(4)
    expected_result[:3, :3] = [
        [0, 1, 0],
        [-1, 0, 0],
        [0, 0, 1],
    ]  # -90 degree rotation around Z-axis
    np.testing.assert_array_almost_equal(matrix_inverse(T), expected_result, decimal=8)

    # Test case 4: Wrong input shape
    with pytest.raises(ValueError):
        matrix_inverse(np.eye(3))

    # Test case 5: Wrong input instance
    with pytest.raises(ValueError):
        matrix_inverse(np.eye(4).tolist())


def test_pose_diff():
    """
    Test the pose_diff function.
    """
    # Test case 1: Identity matrices
    T1 = np.eye(4)
    T2 = np.eye(4)
    expected_result = np.eye(4)
    np.testing.assert_array_equal(pose_diff(T1, T2), expected_result)

    # Test case 2: Translation matrices
    T1 = np.eye(4)
    T1[:3, 3] = [1, 2, 3]
    T2 = np.eye(4)
    T2[:3, 3] = [4, 5, 6]
    expected_result = np.eye(4)
    expected_result[:3, 3] = [3, 3, 3]
    np.testing.assert_array_equal(pose_diff(T1, T2), expected_result)

    # Test case 3: Rotation matrices
    T1 = np.eye(4)
    T1[:3, :3] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]  # 90 degree rotation around Z-axis
    T2 = np.eye(4)
    T2[:3, :3] = [[0, 1, 0], [-1, 0, 0], [0, 0, 1]]  # -90 degree rotation around Z-axis
    expected_result = np.eye(4)
    expected_result[:3, :3] = [
        [-1, 0, 0],
        [0, -1, 0],
        [0, 0, 1],
    ]  # 180 degree rotation around Z-axis
    np.testing.assert_array_almost_equal(pose_diff(T1, T2), expected_result, decimal=8)

    # Test case 4: Not a numpy array
    with pytest.raises(ValueError):
        pose_diff("not a numpy array", np.eye(4))
    with pytest.raises(ValueError):
        pose_diff(np.eye(4), "not a numpy array")

    # Test case 5: Wrong shape
    with pytest.raises(ValueError):
        pose_diff(np.eye(3), np.eye(4))
    with pytest.raises(ValueError):
        pose_diff(np.eye(4), np.eye(3))


def test_adjoint():
    """
    Test the adjoint function.
    """
    # Test case 1: Identity matrix
    T = np.eye(4)
    expected_result = np.eye(6)
    np.testing.assert_array_equal(adjoint(T), expected_result)

    # Test case 2: Translation matrix
    T = np.eye(4)
    T[:3, 3] = [1, 2, 3]
    expected_result = np.eye(6)
    expected_result[3:, :3] = vec_to_so3([1, 2, 3])
    np.testing.assert_array_equal(adjoint(T), expected_result)

    # Test case 3: Rotation matrix
    T = np.eye(4)
    T[:3, :3] = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]  # 90 degree rotation around Z-axis
    expected_result = np.zeros((6, 6))
    expected_result[:3, :3] = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    expected_result[3:, 3:] = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
    np.testing.assert_array_almost_equal(adjoint(T), expected_result)

    # Test case 4: Not a numpy array
    with pytest.raises(ValueError):
        adjoint("not a numpy array")

    # Test case 5: Wrong shape
    with pytest.raises(ValueError):
        adjoint(np.eye(3))


def test_vec_to_se3():
    """
    Test the vec_to_se3 function.
    """
    # Test case 1: Zero vector
    vec = np.zeros(6)
    expected_result = np.zeros((4, 4))
    np.testing.assert_array_equal(vec_to_se3(vec), expected_result)

    # Test case 2: Non-zero vector
    vec = np.array([1, 2, 3, 4, 5, 6])
    expected_result = np.zeros((4, 4))
    expected_result[:3, :3] = vec_to_so3(vec[:3])
    expected_result[:3, 3] = vec[3:]
    np.testing.assert_array_equal(vec_to_se3(vec), expected_result)

    # Test case 3: Not a numpy array or a list
    with pytest.raises(ValueError):
        vec_to_se3("not a numpy array or a list")

    # Test case 4: Wrong number of elements
    with pytest.raises(ValueError):
        vec_to_se3(np.zeros(5))


def test_se3_to_vec():
    """
    Test the se3_to_vec function.
    """
    # Test case 1: Zero matrix
    mat = np.zeros((4, 4))
    expected_result = np.zeros(6)
    np.testing.assert_array_equal(se3_to_vec(mat), expected_result)

    # Test case 2: Non-zero matrix
    mat = np.array([[0, -3, 2, 4], [3, 0, -1, 5], [-2, 1, 0, 6], [0, 0, 0, 0]])
    expected_result = np.array([1, 2, 3, 4, 5, 6])
    np.testing.assert_array_equal(se3_to_vec(mat), expected_result)

    # Test case 3: Not a numpy array
    with pytest.raises(ValueError):
        se3_to_vec("not a numpy array")

    # Test case 4: Wrong shape
    with pytest.raises(ValueError):
        se3_to_vec(np.eye(3))


def test_axis_ang6():
    """
    Test the axis_ang6 function.
    """
    # Test case 1: Zero vector
    vec = np.array([0, 0, 0, 1, 1, 1])
    expected_axis = vec / np.linalg.norm(np.array([1, 1, 1]))
    expected_theta = np.linalg.norm(np.array([1, 1, 1]))
    axis, theta = axis_ang6(vec)
    np.testing.assert_array_equal(axis, expected_axis)
    assert theta == expected_theta

    # Test case 2: Non-zero vector
    vec = np.array([1, 2, 3, 4, 5, 6])
    expected_axis = vec / np.linalg.norm(vec[:3])
    expected_theta = np.linalg.norm(vec[:3])
    axis, theta = axis_ang6(vec)
    np.testing.assert_array_almost_equal(axis, expected_axis, decimal=8)
    assert np.isclose(theta, expected_theta, atol=1e-8)

    # Test case 3: Not a numpy array or a list
    with pytest.raises(ValueError):
        axis_ang6("not a numpy array or a list")

    # Test case 4: Wrong number of elements
    with pytest.raises(ValueError):
        axis_ang6(np.zeros(5))


def test_matrix_exp6():
    """
    Test the matrix_exp6 function.
    """
    # Test case 1: Non-zero matrix
    se3mat = np.array(
        [
            [0, 0, 0, 0],
            [0, 0, -1.57079632, 2.35619449],
            [0, 1.57079632, 0, 2.35619449],
            [0, 0, 0, 0],
        ]
    )
    expected_result = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0],
            [0.0, 1.0, 0.0, 3.0],
            [0, 0, 0, 1],
        ]
    )
    np.testing.assert_array_almost_equal(
        matrix_exp6(se3mat), expected_result, decimal=8
    )

    # Test case 3: Not a numpy array
    with pytest.raises(ValueError):
        matrix_exp6("not a numpy array")

    # Test case 4: Wrong shape
    with pytest.raises(ValueError):
        matrix_exp6(np.eye(3))


def test_matrix_log6():
    """
    Test the matrix_log6 function.
    """
    # Test case 1: Identity matrix
    mat = np.eye(4)
    expected_result = np.zeros((4, 4))
    np.testing.assert_array_equal(matrix_log6(mat), expected_result)

    # Test case 2: Non-identity matrix
    mat = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0],
            [0.0, 1.0, 0.0, 3.0],
            [0, 0, 0, 1],
        ]
    )
    expected_result = np.array(
        [
            [0, 0, 0, 0],
            [0, 0, -1.57079632, 2.35619449],
            [0, 1.57079632, 0, 2.35619449],
            [0, 0, 0, 0],
        ]
    )
    np.testing.assert_array_almost_equal(matrix_log6(mat), expected_result, decimal=8)

    # Test case 3: Not a numpy array
    with pytest.raises(ValueError):
        matrix_log6("not a numpy array")

    # Test case 4: Wrong shape
    with pytest.raises(ValueError):
        matrix_log6(np.eye(3))


def test_project_to_SE3():
    """
    Test the project_to_SE3 function.
    """
    # Test case 1: Non-zero matrix
    mat = np.array(
        [
            [0.675, 0.150, 0.720, 1.2],
            [0.370, 0.771, -0.511, 5.4],
            [-0.630, 0.619, 0.472, 3.6],
            [0.003, 0.002, 0.010, 0.9],
        ]
    )
    expected_result = np.array(
        [
            [0.67901136, 0.14894516, 0.71885945, 1.2],
            [0.37320708, 0.77319584, -0.51272279, 5.4],
            [-0.63218672, 0.61642804, 0.46942137, 3.6],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    np.testing.assert_array_almost_equal(
        project_to_se3(mat), expected_result, decimal=8
    )

    # Test case 3: Not a numpy array
    with pytest.raises(ValueError):
        project_to_se3("not a numpy array")

    # Test case 4: Wrong shape
    with pytest.raises(ValueError):
        project_to_se3(np.eye(3))


def test_distance_to_SE3():
    """
    Test the distance_to_SE3 function.
    """
    # Test case: Given example
    mat = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 0.1, -0.95, 0.0],
            [0.0, 1.0, 0.1, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    expected_result = 0.08835
    assert np.isclose(distance_to_se3(mat), expected_result, atol=1e-5)

    # Test case 1: Not a numpy array
    with pytest.raises(ValueError):
        distance_to_se3("not a numpy array")

    # Test case 2: Wrong shape
    with pytest.raises(ValueError):
        distance_to_se3(np.eye(3))



# import methods from the corresponding modules
from .time import (
    datestr2utime,
    time_interval_indices,
    time_now,
    time_now_utc,
    timestamps_make_uniform,
    utime2datestr,
)

# Get __all__ from the corresponding modules
__all__ = [
    "time_now",
    "time_now_utc",
    "utime2datestr",
    "datestr2utime",
    "time_interval_indices",
    "timestamps_make_uniform",
]



"""
This module provides functions for filtering data.

Functions:
    filter_median: Median filter for multi-dimensional data.
    filter_savgol: Savitsky-Golay filter for multi-dimensional data.
    filter_lowpass: Lowpass filter for multi-dimensional data.

Authors: Sebastián Rodríguez-Martínez and Giancarlo Troni
Contact: srodriguez@mbari.org
"""

from typing import List, Tuple, Union

import numpy as np
from scipy.signal import butter, filtfilt, lfilter, lfiltic, medfilt, savgol_filter

from navlib.math.vmath import resample


def filter_median(
    data: Union[np.ndarray, List[float]],
    time: Union[np.ndarray, List[float]] = None,
    window: int = 13,
    threshold: Union[float, int] = None,
    patch: bool = False,
    remove_outliers: bool = False,
) -> Tuple[np.ndarray, Union[np.ndarray, None]]:
    """
    Median filter for multi-dimensional data. Removes data points that are above
    a threshold, replacing them with nan or interpolated data. The filter is
    applied to each column of the data.

    Args:
        data (Union[np.ndarray, List[float]]): Data to be filtered.
        time (Union[np.ndarray, List[float]]): Time vector.
        window (int, optional): Median filter length. By default 13.
        threshold (Union[float, int], optional): Threshold size defined to remove data. If the
            threashold is not defined STD of the data error is used. By default
            None.
        patch (bool, optional): Patch data. Replace removed data with linear
            interpolation of the points. By default False.
        remove_outliers (bool, optional): Remove outliers. By default False.

    Returns:
        data_filtered (np.ndarray): Filtered data.
        time_filtered (np.ndarray): Filtered time.

    Raises:
        TypeError: If any parameter has an invalid type.
        ValueError: If numeric parameters are non-positive, time is not 1D, or if
            data and time lengths mismatch.

    Examples:
        >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        >>> filtered_data, filtered_time = filter_median(data, window=3)
    """
    # Convert data to numpy array if list
    if isinstance(data, list):
        data = np.array(data)
    if time is not None and isinstance(time, list):
        time = np.array(time)

    # Check the data type
    if not isinstance(data, np.ndarray):
        raise TypeError("Data must be a numpy array or list")
    if time is not None and not isinstance(time, np.ndarray):
        raise TypeError("Time must be a numpy array or list")
    if not isinstance(window, int):
        raise TypeError("Window must be an integer")
    if threshold is not None:
        if not isinstance(threshold, (int, float)):
            raise TypeError("Threshold must be a float or an integer")
        threshold = float(threshold)
    if not isinstance(patch, bool):
        raise TypeError("Patch must be a boolean")
    if not isinstance(remove_outliers, bool):
        raise TypeError("Remove outliers must be a boolean")

    # Make sure data is a 2D array
    if data.ndim == 1:
        data = data[:, np.newaxis]
    # Make sure the time is a 1D array
    if time is not None:
        time = np.squeeze(time)

    # Check values
    if window <= 0:
        raise ValueError("Window must be greater than zero")
    if window % 2 == 0:
        window += 1  # Ensure odd window size for medfilt
    if threshold is not None and threshold <= 0:
        raise ValueError("Threshold must be greater than zero")
    if time is not None:
        if time.ndim != 1:
            raise ValueError("Time must be a 1D array")
        if time.size != data.shape[0]:
            raise ValueError(
                "Time and data must have the same length if single"
                "column or same number of rows if multiple columns"
            )

    # Compute the median filter of the data
    data_filtered = np.zeros(data.shape)
    for ix in range(data.shape[1]):
        data_filtered[:, ix] = medfilt(data[:, ix], window)
    time_filtered = time.copy().astype(float) if time is not None else None

    # Compute residuals and threshold
    if threshold is None:
        residuals = np.abs(data - data_filtered)
        threshold = 3 * np.max(np.std(residuals, axis=0))

        # Identify outliers
        is_bad = np.any(residuals > threshold, axis=1)
        is_good = ~is_bad

        # Handle outliers
        if remove_outliers:
            data_filtered_good = data[is_good]
            time_filtered_good = time[is_good] if time is not None else None
        else:
            data_filtered_good = data_filtered.copy().astype(float)
            data_filtered_good[is_bad] = np.nan
            time_filtered_good = time.copy().astype(float) if time is not None else None
            if time_filtered_good is not None:
                time_filtered_good[is_bad] = np.nan

        # Patch removed data if needed
        if time is not None and patch and remove_outliers:
            # Use external resample function (from your internal library)
            data_filtered = resample(time, time_filtered_good, data_filtered_good)
            time_filtered = time
        else:
            data_filtered = data_filtered_good
            time_filtered = time_filtered_good

    return data_filtered.squeeze(), (
        time_filtered.squeeze() if time_filtered is not None else None
    )


def filter_savgol(
    data: Union[np.ndarray, List[float]],
    time: Union[np.ndarray, List[float]] = None,
    window: int = 10,
    polynomial: int = 2,
    threshold: Union[float, int] = None,
    patch: bool = True,
    remove_outliers: bool = False,
) -> Tuple[np.ndarray, Union[np.ndarray, None]]:
    """
    Savitzky-Golay filter for multi-dimensional data. Removes data points that
    are above a threshold, replacing them with NaN or interpolated data.

    Args:
        data: Data to be filtered. Shape (N,) or (N, D)
        time: Optional time vector (N,)
        window: Smoothing window size. If `time` is provided, interpreted as seconds and converted to samples.
        polynomial: Polynomial order for SG filter. Must be less than window.
        threshold: Threshold to detect outliers (absolute difference from smoothed). If None, defaults to 3*STD of residuals.
        patch: Whether to patch removed data using interpolation.
        remove_outliers: Whether to remove outliers based on the residual threshold.

    Returns:
        data_filtered (np.ndarray): Filtered data.
        time_filtered (np.ndarray): Filtered time.

    Raises:
        TypeError / ValueError: On invalid input types or shapes.
    """
    # Convert to numpy
    if isinstance(data, list):
        data = np.array(data, dtype=float)
    if time is not None and isinstance(time, list):
        time = np.array(time, dtype=float)

    if not isinstance(data, np.ndarray):
        raise TypeError("Data must be a numpy array or list.")
    if time is not None and not isinstance(time, np.ndarray):
        raise TypeError("Time must be a numpy array or list.")
    if not isinstance(window, int) and window <= 0:
        raise ValueError("Window must be a positive integer.")
    if not isinstance(polynomial, int) and polynomial <= 0:
        raise ValueError("Polynomial must be a positive integer.")
    if threshold is not None and (
        not isinstance(threshold, (int, float)) and threshold <= 0
    ):
        raise ValueError("Threshold must be a positive float.")
    if not isinstance(patch, bool):
        raise TypeError("Patch must be a boolean.")
    if not isinstance(remove_outliers, bool):
        raise TypeError("Remove outliers must be a boolean.")

    # Ensure 2D data
    if data.ndim == 1:
        data = data[:, np.newaxis]
    N, D = data.shape

    # Check time shape and match
    if time is not None:
        time = np.squeeze(time)
        if time.ndim != 1:
            raise ValueError("Time must be a 1D array.")
        if time.shape[0] != N:
            raise ValueError("Time and data must have the same number of samples.")

        # Convert smoothing window in seconds to samples
        dt = np.median(np.diff(time))
        window_samples = int(window / dt)
        if window_samples % 2 == 0:
            window_samples += 1
        window = window_samples

    # Final window checks
    if window > N:
        raise ValueError("Window must not exceed number of samples.")
    if window <= polynomial:
        raise ValueError("Window must be greater than polynomial order.")

    # Apply Savitzky-Golay filter
    data_filtered = savgol_filter(
        data, window_length=window, polyorder=polynomial, axis=0
    )
    time_filtered = time.copy().astype(float) if time is not None else None

    # Compute residuals and threshold
    if threshold is not None:
        residuals = np.abs(data - data_filtered)
        threshold = 3 * np.max(np.std(residuals, axis=0))

        # Identify outliers
        is_bad = np.any(residuals > threshold, axis=1)
        is_good = ~is_bad

        # Handle outliers
        if remove_outliers:
            data_filtered_good = data[is_good]
            time_filtered_good = time[is_good] if time is not None else None
        else:
            data_filtered_good = data_filtered.copy().astype(float)
            data_filtered_good[is_bad] = np.nan
            time_filtered_good = time.copy().astype(float) if time is not None else None
            if time_filtered_good is not None:
                time_filtered_good[is_bad] = np.nan

        # Patch removed data if needed
        if time is not None and patch and remove_outliers:
            # Use external resample function (from your internal library)
            data_filtered = resample(time, time_filtered_good, data_filtered_good)
            time_filtered = time
        else:
            data_filtered = data_filtered_good
            time_filtered = time_filtered_good

    return data_filtered.squeeze(), (
        time_filtered.squeeze() if time_filtered is not None else None
    )


def filter_lowpass(
    data: Union[np.ndarray, List[float]],
    sample_freq_hz: Union[int, float] = 1.0,
    cutoff_freq_hz: Union[int, float] = 0.1,
    filter_order: int = 5,
    causality: str = "noncausal",
) -> np.ndarray:
    """
    Lowpass filter for multi-dimensional data for a signal with a given sample
    frequency and cutoff frequency. The filter is applied to each column of the
    data.

    Args:
        data (Union[np.ndarray, List[float]]): Data to be filtered.
        sample_freq_hz (Union[int, float], optional): Sample frequency in Hz. By default 1.0.
        cutoff_freq_hz (Union[int, float], optional): Cutoff frequency in Hz. By default 0.5.
        filter_order (int, optional): Filter order. By default 5.
        causality (str, optional): Causality of the filter. By default "noncausal".

    Returns:
        data_filtered (np.ndarray): Filtered data.

    Raises:
        TypeError: If any parameter has an invalid type.
        ValueError: If numeric parameters are non-positive or if causality is
            invalid.
    Examples:
        >>> data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        >>> filtered_data = filter_lowpass(data)
    """
    # Convert data to numpy array if list
    if isinstance(data, list):
        data = np.array(data)

    # Check the data type
    if not isinstance(data, np.ndarray):
        raise TypeError("Data must be a numpy array or list")
    if not isinstance(sample_freq_hz, (int, float)):
        raise TypeError("Sample frequency must be a float")
    if not isinstance(cutoff_freq_hz, (int, float)):
        raise TypeError("Cutoff frequency must be a float")
    if not isinstance(filter_order, int):
        raise TypeError("Filter order must be an integer")
    if not isinstance(causality, str):
        raise TypeError("Causality must be a string")

    # Make causality lowercase
    causality = causality.lower()

    # Make sure data is a 2D array
    if data.ndim == 1:
        data = data[:, np.newaxis]

    # Check values
    if sample_freq_hz <= 0:
        raise ValueError("Sample frequency must be greater than zero")
    if cutoff_freq_hz <= 0:
        raise ValueError("Cutoff frequency must be greater than zero")
    if filter_order <= 0:
        raise ValueError("Filter order must be greater than zero")
    if causality not in {"causal", "noncausal", "non-causal", "no causal", "acausal"}:
        raise ValueError(
            "Causality must be 'causal', 'noncausal', 'non-causal', 'no causal', or 'acausal'"
        )

    # Filter Design
    causal = True if causality == "causal" else False
    wn = cutoff_freq_hz / (0.5 * sample_freq_hz)

    if wn >= 1 or wn <= 0:
        raise ValueError("The cutoff frequency cannot exceed 1/2 the sample frequency.")

    # Create nth order butterworth filter
    b_matrix, a_matrix = butter(filter_order, wn, btype="low", analog=False)

    # Apply filter to each column of the data
    if causal:
        # Causal filtering. Initialize the filter state using the initial
        # conditions of the filter.
        zi = lfiltic(b_matrix, a_matrix, data[0, :])
        data_filtered, _ = lfilter(
            b_matrix, a_matrix, data, zi=zi[:, np.newaxis], axis=0
        )
    else:
        # Non-causal filtering. Apply the filter forward and backward in time
        # to remove phase shift.
        data_filtered = filtfilt(b_matrix, a_matrix, data, axis=0)

    return data_filtered.squeeze()



# import methods from the corresponding modules
from .log_to_smat import lcmlog_to_dict, msg_getconstants, msg_getfields, parse_and_save
from .parse_lcmlog import (
    dict_to_struct,
    filter_channels_from_data,
    merge_lcm_logs,
    nav_structtype,
    read_log,
    read_logs,
    struct_to_dict,
)

# Get __all__ from the corresponding modules
# log_to_smat module
__all__ = [
    "msg_getfields",
    "msg_getconstants",
    "parse_and_save",
    "lcmlog_to_dict",
]
# parse_lcmlog module
__all__ += [
    "dict_to_struct",
    "filter_channels_from_data",
    "merge_lcm_logs",
    "read_log",
    "read_logs",
    "nav_structtype",
    "struct_to_dict",
]



"""
This module provides functions for working with time.

Functions:
    time_now: Returns the current time in microseconds since the epoch in local time.
    time_now_utc: Returns the current time in microseconds since the epoch in UTC.
    utime2datestr: Converts a Unix timestamp to a date string.
    datestr2utime: Converts a date string to a Unix timestamp.

Authors: Sebastián Rodríguez-Martínez and Giancarlo Troni
Contact: srodriguez@mbari.org
"""
import datetime
from typing import List, Union

import numpy as np


def time_interval_indices(
    time: Union[List[float], np.ndarray],
    t0: Union[int, float],
    tn: Union[int, float],
) -> np.ndarray:
    """
    Given a set of timestamps, return the range of indices that is between t0 and tn.
    This function is useful for filtering timestamps within a specific time interval.

    Args:
        time (Union[List[float], np.ndarray]): Non-saturated values as a list of floats or as a numpy array.
        t0 (Union[int, float]): Lower boundary as a float or int.
        tn (Union[int, float]): Upper boundary as a float or int.

    Returns:
        idx_range (Union[List[float], np.ndarray]): Index range between t0 and tn.

    Raises:
        TypeError: If the input time is not a list or numpy array, or if t0 and tn are not int or float.
        ValueError: If the input time is not 1D, if it is empty, or if tn is not greater than t0.
    """
    # Convert to numpy array
    if isinstance(time, list):
        time = np.array(time)

    # Check input type
    if not isinstance(time, np.ndarray):
        raise TypeError("Input time must be a list or numpy array.")
    if not isinstance(t0, (int, float)):
        raise TypeError("t0 must be an int or float.")
    if not isinstance(tn, (int, float)):
        raise TypeError("tn must be an int or float.")

    # Check time array shape
    time = time.squeeze()
    if time.ndim != 1:
        raise ValueError("Input time must be a 1D array or list.")
    if time.shape[0] == 0:
        raise ValueError("Input time must not be empty.")

    # Check if tn is greater than t0
    if tn <= t0:
        raise ValueError("Upper boundary tn must be greater than lower boundary t0.")

    # Get index range
    idx_1 = np.searchsorted(time, np.max([t0, time[0]]), side="left")
    idx_2 = np.searchsorted(time, np.min([tn, time[-1]]), side="right")

    return np.arange(idx_1, idx_2)


def timestamps_make_uniform(timestamps: Union[List[float], np.ndarray]) -> np.ndarray:
    """
    Given a set of timestamps with non-uniform intervals or non-monotonic order,
    return a new set of timestamps with uniform intervals. This new set of timestamps
    will be evenly spaced between the minimum and maximum timestamps of the input.

    Args:
        timestamps (Union[List[float], np.ndarray]): Non-saturated values as a list of floats or as a numpy array.
    Returns:
        timestamps(np.ndarray): A new set of timestamps with uniform intervals.

    Raises:
        TypeError: If the input timestamps is not a list or numpy array.
        ValueError: If the input timestamps is not 1D or if it is empty.
    """
    # Convert to numpy array
    if isinstance(timestamps, list):
        timestamps = np.array(timestamps)

    # Check input type
    if not isinstance(timestamps, np.ndarray):
        raise TypeError("Input timestamps must be a list or numpy array.")

    # Check timestamps shape
    timestamps = timestamps.squeeze()
    if timestamps.ndim != 1:
        raise ValueError("Input timestamps must be a 1D array or list.")
    if timestamps.shape[0] == 0:
        raise ValueError("Input timestamps must not be empty.")

    # Correct timestamps
    timestamps = np.sort(timestamps)
    timestamps = np.linspace(
        np.min(timestamps), np.max(timestamps), num=timestamps.shape[0]
    )
    return timestamps


def time_now() -> float:
    """
    Returns the current time in microseconds since the epoch in local time.

    Returns:
        unix_ms (float): Current time in microseconds since the epoch.

    Examples:
        >>> time_now()
        1590678950000000.0
    """
    t = datetime.datetime.now()
    time_received = t.timestamp() * 1.00e6
    return time_received


def time_now_utc() -> float:
    """
    Returns the current time in microseconds since the epoch in UTC.

    Returns:
        unix_utc_ms (float): Current time in microseconds since the epoch.

    Examples:
        >>> time_now_utc()
        1590678950000000.0
    """
    t = datetime.datetime.now(datetime.timezone.utc)
    time_received = t.timestamp() * 1.00e6
    return time_received


def utime2datestr(
    utime: Union[List[float], np.ndarray, int, float, np.ndarray],
    datefmt: str = "%Y/%m/%d %H:%M:%S",
) -> Union[List[str], str]:
    """
    Converts unix timestamp in microseconds to a date string.

    Args:
        utime (Union[List[float], np.ndarray, int, float]): Unix timestamp in microseconds.
        datefmt (str, optional): Date format. Defaults to "%Y/%m/%d %H:%M:%S".

    Returns:
        data (Union[List[str], str]): Date string or list of date strings.

    Raises:
        TypeError: If the input utime is not a list, numpy array, int, or float.
        ValueError: If the input utime list is empty or if the date format is incorrect.

    Examples:
        >>> utime2datestr(1590678950000000.0)
        '2020/05/28 12:09:10'
    """
    # Convert to numpy array if it's a list
    if isinstance(utime, list):
        utime = np.array(utime)

    # Check input type
    if not isinstance(utime, (np.ndarray, int, float)):
        raise TypeError("Input utime must be a list, numpy array, int, or float.")

    # Check time array shape
    if isinstance(utime, np.ndarray):
        utime = utime.squeeze()
        if utime.ndim != 1:
            raise ValueError("Input utime must be a 1D array or list.")
        if utime.shape[0] == 0:
            raise ValueError("Input utime must not be empty.")
        if utime.shape[0] == 1:
            utime = utime[0]

    # Convert timestamps
    if isinstance(utime, (int, float)):
        t = datetime.datetime.fromtimestamp(utime / 1.00e6, tz=datetime.timezone.utc)
        return t.strftime(datefmt)
    else:
        t = []
        for i in range(utime.shape[0]):
            t.append(
                datetime.datetime.fromtimestamp(
                    utime[i] / 1.00e6, tz=datetime.timezone.utc
                )
            )
            t[i] = t[i].strftime(datefmt)
        return t


def datestr2utime(
    datestr: Union[str, List[str], np.ndarray],
    datefmt: str = "%Y/%m/%d %H:%M:%S",
) -> Union[float, List[float]]:
    """
    Converts date strings to a Unix timestamp in microseconds.

    Args:
        datestr (Union[str, List[str], np.ndarray]): Date string or list of date strings.
        datefmt (str, optional): Date format. Defaults to "%Y/%m/%d %H:%M:%S".

    Returns:
        unix_ms (Union[float, List[float]]): Unix timestamp in microseconds or list of timestamps.

    Raises:
        TypeError: If the input datestr is not a string, list of strings or numpy array.
        ValueError: If the input datestr list is empty or if the date format is incorrect.

    Examples:
        >>> datestr2utime("2020/05/28 12:09:10")
        1590678950000000.0
    """
    # Check input type
    if not isinstance(datestr, (str, list, np.ndarray)):
        raise TypeError(
            "Input datestr must be a string, list of strings or numpy array."
        )
    if isinstance(datestr, np.ndarray):
        datestr = datestr.tolist()

    # Check list length
    if isinstance(datestr, list):
        if len(datestr) == 0:
            raise ValueError("Input datestr list must not be empty.")
        if len(datestr) == 1:
            datestr = datestr[0]

    # Convert to datetime object
    if isinstance(datestr, str):
        t = datetime.datetime.strptime(datestr, datefmt)
        # Force the timezone to UTC
        t = t.replace(tzinfo=datetime.timezone.utc)
        return t.timestamp() * 1.00e6
    else:
        t = []
        for i in range(len(datestr)):
            t.append(datetime.datetime.strptime(datestr[i], datefmt))
            # Force the timezone to UTC
            t[i] = t[i].replace(tzinfo=datetime.timezone.utc)
            t[i] = t[i].timestamp() * 1.00e6
        return t



"""
Test module for navlib.logging.logging

Authors: Sebastián Rodríguez-Martínez
Contact: srodriguez@mbari.org
"""

import io
import logging
import unittest

from navlib.logging.logging import _NavlibFormatter, setup_logging


class TestNavlibFormatter(unittest.TestCase):
    """Test cases for _NavlibFormatter class."""

    def setUp(self):
        """Set up test fixtures."""
        self.formatter = _NavlibFormatter()

    def test_default_initialization(self):
        """Test formatter initialization with default values."""
        formatter = _NavlibFormatter()
        self.assertEqual(formatter.app_name, "navlib")
        self.assertEqual(formatter.date_format, "%Y/%m/%d %H:%M:%S")

    def test_custom_initialization(self):
        """Test formatter initialization with custom values."""
        formatter = _NavlibFormatter(
            app_name="TestApp", date_format="%H:%M:%S", fmt="Custom format: %(message)s"
        )
        self.assertEqual(formatter.app_name, "TestApp")
        self.assertEqual(formatter.date_format, "%H:%M:%S")

    def test_format_message(self):
        """Test the format method."""
        formatter = _NavlibFormatter(app_name="TestApp")

        # Create a log record
        record = logging.LogRecord(
            name="test.logger",
            level=logging.INFO,
            pathname="",
            lineno=0,
            msg="Test message",
            args=(),
            exc_info=None,
        )

        formatted = formatter.format(record)

        # Check that the formatted message contains expected components
        self.assertIn("[INFO]", formatted)
        self.assertIn("[TestApp]", formatted)
        self.assertIn("[test.logger]", formatted)
        self.assertIn("Test message", formatted)

    def test_format_with_different_log_levels(self):
        """Test formatting with different log levels."""
        formatter = _NavlibFormatter(app_name="TestApp")

        levels = [
            (logging.DEBUG, "DEBUG"),
            (logging.INFO, "INFO"),
            (logging.WARNING, "WARNING"),
            (logging.ERROR, "ERROR"),
            (logging.CRITICAL, "CRITICAL"),
        ]

        for level_int, level_str in levels:
            record = logging.LogRecord(
                name="test",
                level=level_int,
                pathname="",
                lineno=0,
                msg="Test message",
                args=(),
                exc_info=None,
            )
            formatted = formatter.format(record)
            self.assertIn(f"[{level_str}]", formatted)


class TestSetupLogging(unittest.TestCase):
    """Test cases for setup_logging function."""

    def setUp(self):
        """Set up test fixtures."""
        # Clear any existing loggers before each test
        logging.getLogger().handlers.clear()
        for name in list(logging.Logger.manager.loggerDict.keys()):
            if name.startswith("test"):
                logging.getLogger(name).handlers.clear()

    def tearDown(self):
        """Clean up after each test."""
        # Clear loggers after each test
        logging.getLogger().handlers.clear()
        for name in list(logging.Logger.manager.loggerDict.keys()):
            if name.startswith("test"):
                logging.getLogger(name).handlers.clear()

    def test_default_setup(self):
        """Test setup_logging with default parameters."""
        logger = setup_logging()

        # Check logger properties
        self.assertEqual(logger.level, logging.INFO)  # level=2 maps to INFO
        self.assertEqual(len(logger.handlers), 1)
        self.assertIsInstance(logger.handlers[0], logging.StreamHandler)
        self.assertIsInstance(logger.handlers[0].formatter, _NavlibFormatter)

    def test_level_mapping(self):
        """Test the level mapping functionality."""
        level_mappings = [
            (0, logging.CRITICAL),
            (1, logging.DEBUG),
            (2, logging.INFO),
            (3, logging.WARNING),
            (4, logging.ERROR),
            (5, logging.NOTSET),
        ]

        for input_level, expected_level in level_mappings:
            logger = setup_logging(
                level=input_level, logger_name=f"test_level_{input_level}"
            )
            self.assertEqual(logger.level, expected_level)

    def test_invalid_level_type(self):
        """Test that non-integer levels raise ValueError."""
        with self.assertRaises(ValueError) as context:
            setup_logging(level="invalid")
        self.assertIn("Logging level must be an integer", str(context.exception))

    def test_invalid_level_value(self):
        """Test that invalid level values raise ValueError."""
        with self.assertRaises(ValueError) as context:
            setup_logging(level=10)
        self.assertIn("Invalid logging level", str(context.exception))

    def test_custom_app_name(self):
        """Test setup with custom app name."""
        logger = setup_logging(app_name="CustomApp", logger_name="test_custom_app")
        formatter = logger.handlers[0].formatter
        self.assertEqual(formatter.app_name, "CustomApp")

    def test_named_logger(self):
        """Test setup with named logger."""
        logger_name = "test.named.logger"
        logger = setup_logging(logger_name=logger_name)
        self.assertEqual(logger.name, logger_name)

    def test_custom_handler(self):
        """Test setup with custom handler."""
        custom_handler = logging.FileHandler("test.log")
        logger = setup_logging(
            handler=custom_handler, logger_name="test_custom_handler"
        )

        self.assertEqual(len(logger.handlers), 1)
        self.assertIs(logger.handlers[0], custom_handler)

        # Clean up
        custom_handler.close()
        import os

        if os.path.exists("test.log"):
            os.remove("test.log")

    def test_custom_formatter(self):
        """Test setup with custom formatter."""
        custom_formatter = logging.Formatter("Custom: %(message)s")
        logger = setup_logging(
            formatter=custom_formatter, logger_name="test_custom_formatter"
        )

        formatter = logger.handlers[0].formatter
        self.assertIs(formatter, custom_formatter)

    def test_clear_existing_true(self):
        """Test that existing handlers are cleared when clear_existing=True."""
        logger_name = "test_clear_true"

        # Set up logger with initial handler
        initial_logger = setup_logging(logger_name=logger_name)
        initial_handler_count = len(initial_logger.handlers)

        # Set up same logger again with clear_existing=True (default)
        new_logger = setup_logging(logger_name=logger_name, clear_existing=True)

        # Should have same number of handlers (old ones cleared, new one added)
        self.assertEqual(len(new_logger.handlers), initial_handler_count)

    def test_clear_existing_false(self):
        """Test that existing handlers are kept when clear_existing=False."""
        logger_name = "test_clear_false"

        # Set up logger with initial handler
        initial_logger = setup_logging(logger_name=logger_name)
        initial_handler_count = len(initial_logger.handlers)

        # Set up same logger again with clear_existing=False
        new_logger = setup_logging(logger_name=logger_name, clear_existing=False)

        # Should have more handlers (old ones kept, new one added)
        self.assertEqual(len(new_logger.handlers), initial_handler_count + 1)

    def test_logging_output(self):
        """Test that logging actually produces output."""
        # Capture stderr to test logging output
        captured_output = io.StringIO()

        # Create a handler that writes to our captured output
        test_handler = logging.StreamHandler(captured_output)

        logger = setup_logging(
            level=1,  # DEBUG level
            app_name="TestOutput",
            logger_name="test_output",
            handler=test_handler,
        )

        # Log some messages
        logger.debug("Debug message")
        logger.info("Info message")
        logger.warning("Warning message")
        logger.error("Error message")

        # Get the captured output
        output = captured_output.getvalue()

        # Check that all messages are present
        self.assertIn("Debug message", output)
        self.assertIn("Info message", output)
        self.assertIn("Warning message", output)
        self.assertIn("Error message", output)
        self.assertIn("[TestOutput]", output)
        self.assertIn("[test_output]", output)

    def test_return_value(self):
        """Test that setup_logging returns a Logger instance."""
        logger = setup_logging(logger_name="test_return")
        self.assertIsInstance(logger, logging.Logger)


class TestIntegration(unittest.TestCase):
    """Integration tests for the logging module."""

    def setUp(self):
        """Set up test fixtures."""
        # Clear any existing loggers
        logging.getLogger().handlers.clear()

    def tearDown(self):
        """Clean up after tests."""
        logging.getLogger().handlers.clear()

    def test_multiple_loggers(self):
        """Test setting up multiple loggers with different configurations."""
        # Set up different loggers
        nav_logger = setup_logging(
            level=1,  # DEBUG
            app_name="Navigation",
            logger_name="nav",
            clear_existing=False,
        )

        cal_logger = setup_logging(
            level=3,  # WARNING
            app_name="Calibration",
            logger_name="cal",
            clear_existing=False,
        )

        # Test that they work independently
        self.assertEqual(nav_logger.level, logging.DEBUG)
        self.assertEqual(cal_logger.level, logging.WARNING)
        self.assertEqual(nav_logger.name, "nav")
        self.assertEqual(cal_logger.name, "cal")

    def test_logger_hierarchy(self):
        """Test that logger hierarchy works correctly."""
        parent_logger = setup_logging(logger_name="parent", level=1)
        child_logger = setup_logging(
            logger_name="parent.child", level=2, clear_existing=False
        )

        # Both loggers should exist
        self.assertEqual(parent_logger.name, "parent")
        self.assertEqual(child_logger.name, "parent.child")


if __name__ == "__main__":
    unittest.main()



"""
Test module for the Kalman Filter estimation module.

Authors: Sebastián Rodríguez-Martínez
Contact: srodriguez@mbari.org
"""

import inspect

import numpy as np

from navlib import UUV_EKF, kf_lti_discretize, kf_predict, kf_update
from navlib.nav import UUV_EKF as UUV_EKF_relative
from navlib.nav import kf_lti_discretize as kf_lti_discretize_relative
from navlib.nav import kf_predict as kf_predict_relative
from navlib.nav import kf_update as kf_update_relative


def test_imports():
    # Checks that the absolute and relative imports work correctly for functions
    assert (
        kf_lti_discretize_relative.__code__.co_code
        == kf_lti_discretize.__code__.co_code
    )
    assert kf_predict_relative.__code__.co_code == kf_predict.__code__.co_code
    assert kf_update_relative.__code__.co_code == kf_update.__code__.co_code

    # For classes, compare their method code objects
    methods = [
        "initialize_state",
        "predict",
        "correction_position_xy",
        "correction_position_xyz",
        "correction_depth",
        "correction_attitude",
        "correction_angular_rate",
        "correction_velocity",
        "correction_bottom_height",
        "_state_transition_jacobian",
        "_process_noise_covariance",
    ]
    for method in methods:
        rel_method = getattr(UUV_EKF_relative, method)
        abs_method = getattr(UUV_EKF, method)
        assert inspect.getsource(rel_method) == inspect.getsource(abs_method)

    # Compare property names
    rel_props = {
        k for k, v in vars(UUV_EKF_relative).items() if isinstance(v, property)
    }
    abs_props = {k for k, v in vars(UUV_EKF).items() if isinstance(v, property)}
    assert rel_props == abs_props


def test_kf_lti_discretize():
    """
    Test the kf_lti_discretize function.
    """
    # Test case 1: Valid input
    Ac = np.array([[0, 1], [-1, -1]])
    Bc = np.array([[0], [1]])
    Qc = np.array([[1, 0], [0, 1]])
    dt = 0.1
    Ad, Bd, Qd = kf_lti_discretize(Ac, Bc, Qc, dt)
    assert Ad.shape == (2, 2)
    assert Bd.shape == (2, 1)
    assert Qd.shape == (2, 2)

    # Test case 2: List input
    Ad, Bd, Qd = kf_lti_discretize(Ac.tolist(), Bc.tolist(), Qc.tolist(), dt)
    assert Ad.shape == (2, 2)
    assert Bd.shape == (2, 1)
    assert Qd.shape == (2, 2)

    # Test case 3: Default Bc and Qc
    Ad, Bd, Qd = kf_lti_discretize(Ac, dt=dt)
    assert Ad.shape == (2, 2)
    assert Bd.shape == (2, 1)
    assert Qd.shape == (2, 2)

    # Test case 4: Invalid Ac type
    try:
        kf_lti_discretize("invalid_type", Bc, Qc, dt)
    except TypeError as e:
        assert str(e) == "Ac must be a numpy array"
    else:
        assert False

    # Test case 5: Invalid Bc type
    try:
        kf_lti_discretize(Ac, "invalid_type", Qc, dt)
    except TypeError as e:
        assert str(e) == "Bc must be a numpy array"
    else:
        assert False

    # Test case 6: Invalid Qc type
    try:
        kf_lti_discretize(Ac, Bc, "invalid_type", dt)
    except TypeError as e:
        assert str(e) == "Qc must be a numpy array"
    else:
        assert False

    # Test case 7: Invalid dt type
    try:
        kf_lti_discretize(Ac, Bc, Qc, "invalid_type")
    except TypeError as e:
        assert str(e) == "dt must be a number"
    else:
        assert False

    # Test case 8: Invalid Ac shape
    try:
        kf_lti_discretize(np.array([1, 2, 3, 4]), Bc, Qc, dt)
    except ValueError as e:
        assert str(e) == "Ac must be a 2D matrix"
    else:
        assert False

    # Test case 9: Invalid Bc shape
    try:
        kf_lti_discretize(Ac, np.array([1, 2, 3]), Qc, dt)
    except ValueError as e:
        assert str(e) == "Ac and Bc must have the same number of rows"
    else:
        assert False

    # Test case 10: Invalid Qc shape (not square)
    try:
        kf_lti_discretize(Ac, Bc, np.array([[1, 0, 0], [0, 1, 0]]), dt)
    except ValueError as e:
        assert str(e) == "Qc must be a square matrix"
    else:
        assert False

    # Test case 11: Invalid Qc shape (mismatch with Ac)
    try:
        kf_lti_discretize(Ac, Bc, np.array([[1, 0, 0], [0, 1, 0], [0, 1, 0]]), dt)
    except ValueError as e:
        assert str(e) == "Qc must have the same number of rows as Ac"
    else:
        assert False


def test_uuv_ekf_state_consistency():
    # Test that two instances (relative and absolute) behave the same for basic operations
    ekf_abs = UUV_EKF()
    ekf_rel = UUV_EKF_relative()
    # Initialize both
    timestamp = 0.0
    pos = [1.0, 2.0, 3.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.1, 0.2, 0.3]
    ang_vel = [0.01, 0.02, 0.03]
    ekf_abs.initialize_state(timestamp, pos, quat, vel, ang_vel)
    ekf_rel.initialize_state(timestamp, pos, quat, vel, ang_vel)
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)

    # Predict step
    new_timestamp = 0.1
    omega_body = np.array([0.01, 0.02, 0.03])
    ekf_abs.predict(new_timestamp, omega_body)
    ekf_rel.predict(new_timestamp, omega_body)
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)

    # Correction step (position)
    d2_abs, scaled_abs = ekf_abs.correction_position_xy(1.0, 1.5, 2.5)
    d2_rel, scaled_rel = ekf_rel.correction_position_xy(1.0, 1.5, 2.5)
    assert np.isclose(d2_abs, d2_rel)
    assert scaled_abs == scaled_rel
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)


def test_uuv_ekf_predict_and_correction_depth_consistency():
    ekf_abs = UUV_EKF()
    ekf_rel = UUV_EKF_relative()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf_abs.initialize_state(timestamp, pos, quat, vel, ang_vel)
    ekf_rel.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_timestamp = 0.2
    omega_body = np.array([0.0, 0.0, 0.0])
    ekf_abs.predict(new_timestamp, omega_body)
    ekf_rel.predict(new_timestamp, omega_body)
    d2_abs, scaled_abs = ekf_abs.correction_depth(10.0)
    d2_rel, scaled_rel = ekf_rel.correction_depth(10.0)
    assert np.isclose(d2_abs, d2_rel)
    assert scaled_abs == scaled_rel
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)


def test_uuv_ekf_correction_ahrs_consistency():
    ekf_abs = UUV_EKF()
    ekf_rel = UUV_EKF_relative()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf_abs.initialize_state(timestamp, pos, quat, vel, ang_vel)
    ekf_rel.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_quat = [0.0, 0.0, 1.0, 0.0]
    new_quat = np.array(new_quat) / np.linalg.norm(new_quat)
    new_ang_vel = [0.1, 0.2, 0.3]
    ekf_abs.correction_attitude(new_quat)
    ekf_abs.correction_angular_rate(new_ang_vel)
    ekf_rel.correction_attitude(new_quat)
    ekf_rel.correction_angular_rate(new_ang_vel)
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)


def test_uuv_ekf_correction_velocity_consistency():
    ekf_abs = UUV_EKF()
    ekf_rel = UUV_EKF_relative()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf_abs.initialize_state(timestamp, pos, quat, vel, ang_vel)
    ekf_rel.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2_abs, scaled_abs = ekf_abs.correction_velocity([1.0, 0.0, 0.0])
    d2_rel, scaled_rel = ekf_rel.correction_velocity([1.0, 0.0, 0.0])
    assert np.isclose(d2_abs, d2_rel)
    assert scaled_abs == scaled_rel
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)


def test_uuv_ekf_correction_bottom_height_consistency():
    ekf_abs = UUV_EKF(bottom_height=True)
    ekf_rel = UUV_EKF_relative(bottom_height=True)
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf_abs.initialize_state(timestamp, pos, quat, vel, ang_vel, bottom_height=0.0)
    ekf_rel.initialize_state(timestamp, pos, quat, vel, ang_vel, bottom_height=0.0)
    d2_abs, scaled_abs = ekf_abs.correction_bottom_height(10.0)
    d2_rel, scaled_rel = ekf_rel.correction_bottom_height(10.0)
    assert np.isclose(d2_abs, d2_rel)
    assert scaled_abs == scaled_rel
    np.testing.assert_allclose(ekf_abs.current_state, ekf_rel.current_state)
    np.testing.assert_allclose(ekf_abs.current_covariance, ekf_rel.current_covariance)


def test_kf_predict():
    """
    Test the kf_predict function.
    """
    # Test case 1: Valid input
    x = np.array([1, 2])
    P = np.array([[1, 0], [0, 1]])
    A = np.array([[1, 1], [0, 1]])
    Q = np.array([[0.1, 0], [0, 0.1]])
    B = np.array([[0.5], [1]])
    u = np.array([1])
    x_pred, P_pred = kf_predict(x, P, A, Q, B, u)
    assert x_pred.shape == (2,)
    assert P_pred.shape == (2, 2)

    # Test case 2: List input
    x_pred, P_pred = kf_predict(
        x.tolist(), P.tolist(), A.tolist(), Q.tolist(), B.tolist(), u.tolist()
    )
    assert x_pred.shape == (2,)
    assert P_pred.shape == (2, 2)

    # Test case 3: Default B and u
    x_pred, P_pred = kf_predict(x, P, A, Q)
    assert x_pred.shape == (2,)
    assert P_pred.shape == (2, 2)

    # Test case 4: Invalid x type
    try:
        kf_predict("invalid_type", P, A, Q, B, u)
    except TypeError as e:
        assert str(e) == "x must be a numpy array"
    else:
        assert False

    # Test case 5: Invalid P type
    try:
        kf_predict(x, "invalid_type", A, Q, B, u)
    except TypeError as e:
        assert str(e) == "P must be a numpy array"
    else:
        assert False

    # Test case 6: Invalid A type
    try:
        kf_predict(x, P, "invalid_type", Q, B, u)
    except TypeError as e:
        assert str(e) == "A must be a numpy array"
    else:
        assert False

    # Test case 7: Invalid Q type
    try:
        kf_predict(x, P, A, "invalid_type", B, u)
    except TypeError as e:
        assert str(e) == "Q must be a numpy array"
    else:
        assert False

    # Test case 8: Invalid B type
    try:
        kf_predict(x, P, A, Q, "invalid_type", u)
    except TypeError as e:
        assert str(e) == "B must be a numpy array"
    else:
        assert False

    # Test case 9: Invalid u type
    try:
        kf_predict(x, P, A, Q, B, "invalid_type")
    except TypeError as e:
        assert str(e) == "u must be a numpy array"
    else:
        assert False

    # Test case 10: Invalid x shape
    try:
        kf_predict(np.array([1, 2, 3]), P, A, Q, B, u)
    except ValueError as e:
        assert str(e) == "x and A must have the same number of rows"
    else:
        assert False

    # Test case 11: Invalid P shape
    try:
        kf_predict(x, np.array([1, 2, 3]), A, Q, B, u)
    except ValueError as e:
        assert str(e) == "P must be a square matrix"
    else:
        assert False

    # Test case 12: Invalid A shape
    try:
        kf_predict(x, P, np.array([1, 2, 3]), Q, B, u)
    except ValueError as e:
        assert str(e) == "A must be a 2D matrix"
    else:
        assert False

    # Test case 13: Invalid Q shape
    try:
        kf_predict(x, P, A, np.array([1, 2, 3]), B, u)
    except ValueError as e:
        assert str(e) == "Q must be a square matrix"
    else:
        assert False

    # Test case 14: Invalid B shape
    try:
        kf_predict(x, P, A, Q, np.array([1, 2, 3]), u)
    except ValueError as e:
        assert str(e) == "B must be a 2D matrix"
    else:
        assert False

    # Test case 15: Invalid u shape
    try:
        kf_predict(x, P, A, Q, B, np.array([1, 2, 3]))
    except ValueError as e:
        assert (
            str(e)
            == "The number of columns in B must be equal to the number of rows in u"
        )
    else:
        assert False


def test_kf_update():
    """
    Test the kf_update function.
    """
    # Test case 1: Valid input
    x = np.array([1, 2])
    P = np.array([[1, 0], [0, 1]])
    y = np.array([1, 2])
    H = np.array([[1, 0], [0, 1]])
    R = np.array([[0.1, 0], [0, 0.1]])
    x_upd, P_upd, K, dy, S = kf_update(x, P, y, H, R)
    assert x_upd.shape == (2,)
    assert P_upd.shape == (2, 2)
    assert K.shape == (2, 2)
    assert dy.shape == (2,)
    assert S.shape == (2, 2)

    # Test case 2: List input
    x_upd, P_upd, K, dy, S = kf_update(
        x.tolist(), P.tolist(), y.tolist(), H.tolist(), R.tolist()
    )
    assert x_upd.shape == (2,)
    assert P_upd.shape == (2, 2)
    assert K.shape == (2, 2)
    assert dy.shape == (2,)
    assert S.shape == (2, 2)

    # Test case 3: Invalid x type
    try:
        kf_update("invalid_type", P, y, H, R)
    except TypeError as e:
        assert str(e) == "x must be a numpy array"
    else:
        assert False

    # Test case 4: Invalid P type
    try:
        kf_update(x, "invalid_type", y, H, R)
    except TypeError as e:
        assert str(e) == "P must be a numpy array"
    else:
        assert False

    # Test case 5: Invalid y type
    try:
        kf_update(x, P, "invalid_type", H, R)
    except TypeError as e:
        assert str(e) == "y must be a numpy array"
    else:
        assert False

    # Test case 6: Invalid H type
    try:
        kf_update(x, P, y, "invalid_type", R)
    except TypeError as e:
        assert str(e) == "H must be a numpy array"
    else:
        assert False

    # Test case 7: Invalid R type
    try:
        kf_update(x, P, y, H, "invalid_type")
    except TypeError as e:
        assert str(e) == "R must be a numpy array"
    else:
        assert False

    # Test case 10: Invalid P shape
    try:
        kf_update(x, np.array([1, 2, 3]), y, H, R)
    except ValueError as e:
        assert str(e) == "P must be a square matrix"
    else:
        assert False

    # Test case 11: Invalid H shape
    try:
        kf_update(x, P, y, np.array([1, 2, 3]), R)
    except ValueError as e:
        assert str(e) == "H must be a 2D matrix"
    else:
        assert False

    # Test case 12: Invalid R shape
    try:
        kf_update(x, P, y, H, np.array([1, 2, 3]))
    except ValueError as e:
        assert str(e) == "R must be a square matrix"
    else:
        assert False

    # Test case 13: Dimension mismatch between x and P
    try:
        kf_update(np.array([1, 2, 3]), np.array([[1, 0], [0, 1]]), y, H, R)
    except ValueError as e:
        assert str(e) == "P must have the same number of rows as x"
    else:
        assert False


def test_uuv_ekf_init_and_properties():
    # Test default initialization
    ekf = UUV_EKF()
    assert ekf.state_dim == 13
    assert ekf.current_state.shape == (14,)
    assert ekf.current_covariance.shape == (13, 13)
    assert ekf.H_position.shape == (2, 13)
    assert ekf.R_position.shape == (2, 2)
    assert ekf.H_depth.shape == (1, 13)
    assert ekf.R_depth.shape == (1, 1)
    assert ekf.H_attitude.shape == (3, 13)
    assert ekf.R_attitude.shape == (3, 3)
    assert ekf.H_angrate.shape == (3, 13)
    assert ekf.R_angrate.shape == (3, 3)
    assert ekf.H_vel.shape == (3, 13)
    assert ekf.R_vel.shape == (3, 3)
    assert ekf.H_bottom_height.shape == (1, 13)
    assert ekf.R_bottom_height.shape == (1, 1)

    # Test with bottom_height=True
    ekf2 = UUV_EKF(bottom_height=True)
    assert ekf2.state_dim == 14
    assert ekf2.current_state.shape == (15,)
    assert ekf2.current_covariance.shape == (14, 14)
    assert ekf2.H_position.shape == (2, 14)
    assert ekf2.H_depth.shape == (1, 14)
    assert ekf2.H_attitude.shape == (3, 14)
    assert ekf2.H_angrate.shape == (3, 14)
    assert ekf2.H_vel.shape == (3, 14)
    assert ekf2.H_bottom_height.shape == (1, 14)


def test_uuv_ekf_initialize_state_and_setters():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [1.0, 2.0, 3.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.1, 0.2, 0.3]
    ang_vel = [0.01, 0.02, 0.03]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    np.testing.assert_allclose(ekf.current_state[0], timestamp)
    np.testing.assert_allclose(ekf.current_state[1:4], pos)
    np.testing.assert_allclose(ekf.current_state[4:8], quat)
    np.testing.assert_allclose(ekf.current_state[8:11], vel)
    np.testing.assert_allclose(ekf.current_state[11:14], ang_vel)


def test_uuv_ekf_predict():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [1.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_timestamp = 1.0
    omega_body = np.array([0.0, 0.0, 0.0])
    ekf.predict(new_timestamp, omega_body)
    # Position should have advanced by velocity * dt
    np.testing.assert_allclose(
        ekf.current_state[1:4],
        np.array(pos) + np.array(vel) * (new_timestamp - timestamp),
    )
    # Quaternion should remain unchanged for zero angular velocity
    np.testing.assert_allclose(ekf.current_state[4:8], quat, atol=1e-6)


def test_uuv_ekf_correction_position_xy():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_position_xy(1.0, 2.0)
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, position should move closer to measurement
    assert np.abs(ekf.current_state[1] - 1.0) < 1.0
    assert np.abs(ekf.current_state[2] - 2.0) < 2.0

    # Custom covariance for position
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_position_xy(
        1.0, 2.0, measurement_covariance=np.eye(2) * 10.0
    )
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, position should move closer to measurement
    assert np.abs(ekf.current_state[1] - 1.0) < 1.0
    assert np.abs(ekf.current_state[2] - 2.0) < 2.0


def test_uuv_ekf_correction_position_xyz():
    ekf = UUV_EKF(covariance_depth=10.0)
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_position_xyz(1.0, 2.0, 1.0)
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, position should move closer to measurement
    assert np.abs(ekf.current_state[1] - 1.0) < 1.0
    assert np.abs(ekf.current_state[2] - 2.0) < 2.0
    assert np.abs(ekf.current_state[3] - 1.0) < 1.0

    # Custom covariance for position
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_position_xyz(
        1.0, 2.0, 1.0, measurement_covariance=np.eye(3) * 10.0
    )
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, position should move closer to measurement
    assert np.abs(ekf.current_state[1] - 1.0) < 1.0
    assert np.abs(ekf.current_state[2] - 2.0) < 2.0
    assert np.abs(ekf.current_state[3] - 1.0) < 1.0


def test_uuv_ekf_correction_depth():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    timestamp = 1.0
    d2, scaled = ekf.correction_depth(5.0)
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, depth should move closer to measurement
    assert np.abs(ekf.current_state[3] - 5.0) < 5.0

    # Custom covariance for depth
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    timestamp = 1.0
    d2, scaled = ekf.correction_depth(5.0, measurement_covariance=np.eye(1) * 0.001)
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, depth should move closer to measurement
    assert np.abs(ekf.current_state[3] - 5.0) < 5.0


def test_uuv_ekf_correction_attitude():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_quat = [0.0, 0.0, 1.0, 0.0]
    new_quat = np.array(new_quat) / np.linalg.norm(new_quat)
    ekf.correction_attitude(new_quat)
    # Quaternion should be normalized
    np.testing.assert_allclose(np.linalg.norm(ekf.current_state[4:8]), 1.0, atol=1e-6)

    # Custom covariance for attitude
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_quat = [0.0, 0.0, 1.0, 0.0]
    new_quat = np.array(new_quat) / np.linalg.norm(new_quat)
    ekf.correction_attitude(new_quat, measurement_covariance=np.eye(3) * 0.001)
    # Quaternion should be normalized
    np.testing.assert_allclose(np.linalg.norm(ekf.current_state[4:8]), 1.0, atol=1e-6)


def test_uuv_ekf_correction_angular_rate():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_ang_vel = [0.1, 0.2, 0.3]
    ekf.correction_angular_rate(new_ang_vel)
    # Quaternion should be normalized
    np.testing.assert_allclose(np.linalg.norm(ekf.current_state[4:8]), 1.0, atol=1e-6)

    # Custom covariance for angular rate
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    new_ang_vel = [0.1, 0.2, 0.3]
    ekf.correction_angular_rate(new_ang_vel, measurement_covariance=np.eye(3) * 0.001)
    # Quaternion should be normalized
    np.testing.assert_allclose(np.linalg.norm(ekf.current_state[4:8]), 1.0, atol=1e-6)


def test_uuv_ekf_correction_velocity():
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_velocity([1.0, 0.0, 0.0])
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, velocity should move closer to measurement
    assert np.abs(ekf.current_state[8] - 1.0) < 1.0

    # Custom covariance for velocity
    ekf = UUV_EKF()
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel)
    d2, scaled = ekf.correction_velocity(
        [1.0, 0.0, 0.0], measurement_covariance=np.eye(3) * 0.001
    )
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, velocity should move closer to measurement
    assert np.abs(ekf.current_state[8] - 1.0) < 1.0


def test_uuv_ekf_correction_bottom_height():
    ekf = UUV_EKF(bottom_height=True)
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel, bottom_height=0.0)
    d2, scaled = ekf.correction_bottom_height(10.0)
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, bottom height should move closer to measurement
    assert np.abs(ekf.current_state[-1] - 10.0) < 10.0

    # Custom covariance for bottom height
    ekf = UUV_EKF(bottom_height=True)
    timestamp = 0.0
    pos = [0.0, 0.0, 0.0]
    quat = [0.0, 0.0, 0.0, 1.0]
    vel = [0.0, 0.0, 0.0]
    ang_vel = [0.0, 0.0, 0.0]
    ekf.initialize_state(timestamp, pos, quat, vel, ang_vel, bottom_height=0.0)
    d2, scaled = ekf.correction_bottom_height(
        10.0, measurement_covariance=np.eye(1) * 0.001
    )
    assert isinstance(d2, float)
    assert isinstance(scaled, bool)
    # After correction, bottom height should move closer to measurement
    assert np.abs(ekf.current_state[-1] - 10.0) < 10.0
